package ru.dbotthepony.kbox2d.api

import ru.dbotthepony.kvector.vector.ndouble.Vector2d

/**
 * Profiling data. Times are in milliseconds.
 */
interface IProfileData {
	val step: Long
	val collide: Long
	val solve: Long
	val solveInit: Long
	val solveVelocity: Long
	val solvePosition: Long
	val broadphase: Long
	val solveTOI: Long
	val integratePositions: Long
}

/**
 * Profiling data. Times are in nanoseconds.
 */
data class ProfileSnapshot(
	override val step: Long,
	override val collide: Long,
	override val solve: Long,
	override val solveInit: Long,
	override val solveVelocity: Long,
	override val solvePosition: Long,
	override val broadphase: Long,
	override val solveTOI: Long,
	override val integratePositions: Long,
) : IProfileData

/**
 * Profiling data. Times are in nanoseconds.
 */
internal data class ProfileData(
	override var step: Long = 0L,
	override var collide: Long = 0L,
	override var solve: Long = 0L,
	override var solveInit: Long = 0L,
	override var solveVelocity: Long = 0L,
	override var solvePosition: Long = 0L,
	override var broadphase: Long = 0L,
	override var solveTOI: Long = 0L,
	override var integratePositions: Long = 0L,
) : IProfileData {
	fun snapshot(): ProfileSnapshot {
		return ProfileSnapshot(
			step = step,
			collide = collide,
			solve = solve,
			solveInit = solveInit,
			solveVelocity = solveVelocity,
			solvePosition = solvePosition,
			broadphase = broadphase,
			solveTOI = solveTOI,
			integratePositions = integratePositions,
		)
	}
}

/**
 * This is an internal structure.
 */
data class B2TimeStep(
	var dt: Double,
	var inv_dt: Double,
	var dtRatio: Double,
	var velocityIterations: Int,
	var positionIterations: Int,
	var warmStarting: Boolean,
)

/**
 * This is an internal structure.
 */
internal class B2Position(
	c: Vector2d = Vector2d.ZERO,
	a: Double = 0.0,
) {
	var c: Vector2d = c
		set(value) {
			if (!value.isFinite) {
				throw IllegalArgumentException("Tried to set illegal position $value")
			}

			field = value
		}

	var a: Double = a
		set(value) {
			if (!value.isFinite()) {
				throw IllegalArgumentException("Tried to set non-finite angle $value")
			}

			if (value.isNaN()) {
				throw IllegalArgumentException("Tried to set NaN angle")
			}

			field = value
		}

	init {
		// KBox2D: trigger sanity checks at least once
		this.c = c
		this.a = a
	}
}

/**
 * This is an internal structure.
 */
internal class B2Velocity(
	v: Vector2d = Vector2d.ZERO,
	w: Double = 0.0,
) {
	var v: Vector2d = v
		set(value) {
			if (!value.isFinite) {
				throw IllegalArgumentException("Tried to set illegal velocity $value")
			}

			field = value
		}

	var w: Double = w
		set(value) {
			if (!value.isFinite()) {
				throw IllegalArgumentException("Tried to set non-finite angular angle $value")
			}

			if (value.isNaN()) {
				throw IllegalArgumentException("Tried to set NaN angular angle")
			}

			field = value
		}

	init {
		// KBox2D: trigger sanity checks at least once
		this.v = v
		this.w = w
	}
}

/**
 * Solver Data
 */
internal class B2SolverData(
	var step: B2TimeStep,
	var positions: List<B2Position>,
	var velocities: List<B2Velocity>
)
