package ru.dbotthepony.kbox2d.collision.handler

import ru.dbotthepony.kbox2d.api.*
import ru.dbotthepony.kbox2d.api.b2Dot
import ru.dbotthepony.kbox2d.api.b2Mul
import ru.dbotthepony.kbox2d.api.b2MulT
import ru.dbotthepony.kbox2d.collision.b2ClipSegmentToLine
import ru.dbotthepony.kbox2d.collision.shapes.PolygonShape
import ru.dbotthepony.kvector.vector.ndouble.times
import java.util.*
import kotlin.collections.ArrayList

internal data class FindMaxSeparationResult(
	val edgeIndex: Int,
	val maxSeparation: Double
)

// Find the max separation between poly1 and poly2 using edge normals from poly1.
internal fun b2FindMaxSeparation(
	poly1: PolygonShape,
	xf1: Transform,
	poly2: PolygonShape,
	xf2: Transform
): FindMaxSeparationResult {
	val count1 = poly1.vertices.size
	val count2 = poly2.vertices.size
	val n1s = poly1.normals
	val v1s = poly1.vertices
	val v2s = poly2.vertices

	var bestIndex = 0
	var maxSeparation = -Double.MAX_VALUE

	val xf = b2MulT(xf2, xf1)

	for (i in 0 until count1) {
		// Get poly1 normal in frame2.
		val n = b2Mul(xf.q, n1s[i])
		val v1 = b2Mul(xf, v1s[i])

		// Find deepest point for normal i.
		var si = Double.MAX_VALUE
		for (j in 0 until count2) {
			val sij = b2Dot(n, v2s[j] - v1)

			if (sij < si) {
				si = sij
			}
		}

		if (si > maxSeparation) {
			maxSeparation = si
			bestIndex = i
		}
	}

	return FindMaxSeparationResult(
		edgeIndex = bestIndex,
		maxSeparation = maxSeparation
	)
}

internal fun b2FindIncidentEdge(
	poly1: PolygonShape,
	xf1: Transform,
	edge1: Int,
	poly2: PolygonShape,
	xf2: Transform
): Array<ClipVertex> {
	val normals1 = poly1.normals
	val count2 = poly2.count
	val vertices2 = poly2.vertices
	val normals2 = poly2.normals

	require(edge1 in 0 until poly1.count) { "$edge1 is not in range of 0 until ${poly1.count}" }

	// Get the normal of the reference edge in poly2's frame.
	val normal1 = b2MulT(xf2.q, b2Mul(xf1.q, normals1[edge1]))

	// Find the incident edge on poly2.
	var index = 0
	var minDot = Double.MAX_VALUE

	for (i in 0 until count2) {
		val dot = b2Dot(normal1, normals2[i])
		if (dot < minDot) {
			minDot = dot
			index = i
		}
	}

	// Build the clip vertices for the incident edge.
	val i1 = index
	val i2 = if (i1 + 1 < count2) i1 + 1 else 0

	val c0 = ClipVertex(
		v = b2Mul(xf2, vertices2[i1]),
		id = ContactID(
			cf = ContactFeature(
				indexA = edge1,
				indexB = i1,
				typeA = ContactFeature.Type.FACE,
				typeB = ContactFeature.Type.VERTEX,
			)
		)
	)

	val c1 = ClipVertex(
		v = b2Mul(xf2, vertices2[i2]),
		id = ContactID(
			cf = ContactFeature(
				indexA = edge1,
				indexB = i2,
				typeA = ContactFeature.Type.FACE,
				typeB = ContactFeature.Type.VERTEX
			)
		)
	)

	return arrayOf(c0, c1)
}

// Find edge normal of max separation on A - return if separating axis is found
// Find edge normal of max separation on B - return if separation axis is found
// Choose reference edge as min(minA, minB)
// Find incident edge
// Clip

private const val k_tol = 0.1 * b2_linearSlop

// The normal points from 1 to 2
internal fun b2CollidePolygons(
	polyA: PolygonShape,
	xfA: Transform,
	polyB: PolygonShape,
	xfB: Transform
): Manifold {
	val totalRadius = polyA.radius + polyB.radius

	val (edgeA, separationA) = b2FindMaxSeparation(polyA, xfA, polyB, xfB)

	if (separationA > totalRadius) {
		return Manifold.EMPTY
	}

	val (edgeB, separationB) = b2FindMaxSeparation(polyB, xfB, polyA, xfA)

	if (separationB > totalRadius) {
		return Manifold.EMPTY
	}

	val poly1: PolygonShape // reference polygon
	val poly2: PolygonShape // incident polygon
	val xf1: Transform
	val xf2: Transform
	val edge1: Int // reference edge
	val flip: Boolean

	val type: Manifold.Type

	if (separationB > separationA + k_tol) {
		poly1 = polyB
		poly2 = polyA
		xf1 = xfB
		xf2 = xfA
		edge1 = edgeB
		type = Manifold.Type.FACE_B
		flip = true
	} else {
		poly1 = polyA
		poly2 = polyB
		xf1 = xfA
		xf2 = xfB
		edge1 = edgeA
		type = Manifold.Type.FACE_A
		flip = false
	}

	val incidentEdge = b2FindIncidentEdge(poly1, xf1, edge1, poly2, xf2)
	val count1 = poly1.count
	val vertices1 = poly1.vertices

	val iv1 = edge1
	val iv2 = if (edge1 + 1 < count1) edge1 + 1 else 0

	var v11 = vertices1[iv1]
	var v12 = vertices1[iv2]

	val localTangent = (v12 - v11).normalized

	val localNormal = b2Cross(localTangent, 1.0);
	val planePoint = 0.5 * (v11 + v12)

	val tangent = b2Mul(xf1.q, localTangent)
	val normal = b2Cross(tangent, 1.0)

	v11 = b2Mul(xf1, v11)
	v12 = b2Mul(xf1, v12)

	// Face offset.
	val frontOffset = b2Dot(normal, v11)

	// Side offsets, extended by polytope skin thickness.
	val sideOffset1 = -b2Dot(tangent, v11) + totalRadius
	val sideOffset2 = b2Dot(tangent, v12) + totalRadius

	// Clip incident edge against extruded edge1 side edges.
	val clipPoints1 = b2ClipSegmentToLine(incidentEdge, -tangent, sideOffset1, iv1)

	if (clipPoints1.size < 2)
		return Manifold(type = type)

	// Clip to negative box side 1
	val clipPoints2 = b2ClipSegmentToLine(clipPoints1, tangent, sideOffset2, iv2)

	if (clipPoints2.size < 2)
		return Manifold(type = type)

	// Now clipPoints2 contains the clipped points.
	val points = ArrayList<ManifoldPoint>()

	for (i in 0 until b2_maxManifoldPoints) {
		val separation = b2Dot(normal, clipPoints2[i].v) - frontOffset

		if (separation <= totalRadius) {
			val cp = ManifoldPoint(
				localPoint = b2MulT(xf2, clipPoints2[i].v),
				id = clipPoints2[i].id
			)

			if (flip) {
				// Swap features
				val cf = cp.id.cf
				cp.id.cf.indexA = cf.indexB
				cp.id.cf.indexB = cf.indexA
				cp.id.cf.typeA = cf.typeB
				cp.id.cf.typeB = cf.typeA
			}

			points.add(cp)
		}
	}

	return Manifold(
		localPoint = planePoint,
		localNormal = localNormal,
		points = Collections.unmodifiableList(points),
		type = type
	)
}
