package ru.dbotthepony.kbox2d.dynamics

import ru.dbotthepony.kbox2d.api.*
import ru.dbotthepony.kvector.vector.ndouble.Vector2d

class Body(def: BodyDef, world: B2World) {
	private var _world: B2World? = world

	/**
	 * Get the parent world of this body.
	 */
	val world: B2World
		get() = _world ?: throw IllegalStateException("Tried to use removed body")

	internal var flags: Int = 0
	internal var isOnIsland = false
	internal var islandIndex: Int = 0

	internal fun unlink() {
		_world = null
	}

	init {
		def.validate()

		if (def.bullet) {
			flags = BodyFlags.BULLET.or(flags)
		}

		if (def.fixedRotation) {
			flags = BodyFlags.FIXED_ROTATION.or(flags)
		}

		if (def.allowSleep) {
			flags = BodyFlags.AUTO_SLEEP.or(flags)
		}

		if (def.awake && def.type != BodyType.STATIC) {
			flags = BodyFlags.AWAKE.or(flags)
		}

		if (def.enabled) {
			flags = BodyFlags.ENABLED.or(flags)
		}
	}

	internal val sweep: Sweep = Sweep()

	/**
	 * Get the body transform for the body's origin.
	 * @return the world transform of the body's origin.
	 */
	val transform: Transform = Transform(def.position, def.angle)
	val xf by this::transform

	/**
	 * Get the world body origin position.
	 * @return the world position of the body's origin.
	 */
	val position: Vector2d
		get() = transform.position

	/**
	 * Get the angle in radians.
	 * @return the current world rotation angle in radians.
	 */
	val angle: Double
		get() = sweep.a

	val userData: Any? = def.userData

	internal var sleepTime: Double = 0.0
	internal var torque: Double = 0.0
	internal var force: Vector2d = Vector2d.ZERO

	/**
	 * The linear velocity of the body's origin in world co-ordinates.
	 */
	var linearVelocity: Vector2d = def.linearVelocity
		set(value) {
			if (type == BodyType.STATIC)
				return

			if (value.dot(value) > 0.0)
				isAwake = true

			field = value
		}

	/**
	 * The angular velocity of the body.
	 */
	var angularVelocity: Double = def.angularVelocity
		set(value) {
			if (type == BodyType.STATIC)
				return

			if (value * value > 0.0)
				isAwake = true

			field = value
		}

	var fixtureList: Fixture? = null
		protected set

	var fixtureCount: Int = 0
		private set

	/**
	 * Get the list of all joints attached to this body.
	 *
	 * Kotlin: This is not a list, but, as in C++ impl, a custom
	 * linked list.
	 */
	var jointList: JointEdge? = null
		internal set

	/**
	 * Get the list of all contacts attached to this body.
	 * @warning this list changes during the time step and you may
	 * miss some collisions if you don't use b2ContactListener.
	 *
	 * Kotlin: This is not a list, but, as in C++ impl, a custom
	 * linked list.
	 */
	var contactEdge: ContactEdge? = null
		internal set

	val contactEdgeIterator: Iterator<ContactEdge> get() {
		return object : Iterator<ContactEdge> {
			private var node = contactEdge

			override fun hasNext(): Boolean {
				return node != null
			}

			override fun next(): ContactEdge {
				val old = node!!
				node = old.next
				return old
			}
		}
	}

	val jointIterator: Iterator<JointEdge> get() {
		return object : Iterator<JointEdge> {
			private var node = jointList

			override fun hasNext(): Boolean {
				return node != null
			}

			override fun next(): JointEdge {
				val old = node!!
				node = old.next
				return old
			}
		}
	}

	var next: Body? = null
		internal set
	var prev: Body? = null
		internal set

	val fixtureIterator: Iterator<Fixture> get() {
		return object : Iterator<Fixture> {
			private var node = fixtureList

			override fun hasNext(): Boolean {
				return node != null
			}

			override fun next(): Fixture {
				val old = node!!
				node = old.next
				return old
			}
		}
	}

	var linearDamping: Double = def.linearDamping
	var angularDamping: Double = def.angularDamping
	var gravityScale: Double = def.gravityScale

	/**
	 * Get the total mass of the body.
	 * @return the mass, usually in kilograms (kg).
	 */
	var mass: Double = 0.0
		private set

	internal var invMass: Double = 0.0

	/**
	 * Rotational inertia about the center of mass.
 	 */
	internal var rotInertia: Double = 0.0

	/**
	 * Rotational inertia about the center of mass.
	 */
	internal var rotInertiaInv: Double = 0.0

	internal var I by this::rotInertia
	internal var invI by this::rotInertiaInv

	/**
	 * Should this body be treated like a bullet for continuous collision detection?
	 */
	var isBullet: Boolean
		get() = BodyFlags.BULLET.isit(flags)
		set(value) { flags = BodyFlags.BULLET.update(flags, value) }

	/**
	 * Set the sleep state of the body. A sleeping body has very
	 * low CPU cost.
	 * @param flag set to true to wake the body, false to put it to sleep.
	 */
	var isAwake: Boolean
		get() = BodyFlags.AWAKE.isit(flags)
		set(value) {
			if (type == BodyType.STATIC || BodyFlags.AWAKE.isit(flags) == value)
				return

			if (value) {
				flags = flags or BodyFlags.AWAKE.bitmask
				sleepTime = 0.0
			} else {
				flags = flags and BodyFlags.AWAKE.bitmask.inv()
				sleepTime = 0.0
				linearVelocity = Vector2d.ZERO
				angularVelocity = 0.0
				force = Vector2d.ZERO
				torque = 0.0
			}
		}

	/**
	 * Allow a body to be disabled. A disabled body is not simulated and cannot
	 * be collided with or woken up.
	 *
	 * If you pass a flag of true, all fixtures will be added to the broad-phase.
	 *
	 * If you pass a flag of false, all fixtures will be removed from the
	 * broad-phase and all contacts will be destroyed.
	 *
	 * Fixtures and joints are otherwise unaffected. You may continue
	 * to create/destroy fixtures and joints on disabled bodies.
	 *
	 * Fixtures on a disabled body are implicitly disabled and will
	 * not participate in collisions, ray-casts, or queries.
	 *
	 * Joints connected to a disabled body are implicitly disabled.
	 *
	 * An diabled body is still owned by a b2World object and remains
	 * in the body list.
	 */
	var isEnabled: Boolean
		get() = BodyFlags.ENABLED.isit(flags)
		set(value) {
			if (world.isLocked)
				throw ConcurrentModificationException()

			if (value == isEnabled)
				return

			if (value) {
				flags = BodyFlags.ENABLED.not(flags)

				// Create all proxies.
				val broadPhase = world.contactManager.broadPhase

				for (fixture in fixtureIterator) {
					(fixture as Fixture).createProxies(broadPhase, transform)
				}

				world.notifyNewContacts()
			} else {
				flags = BodyFlags.ENABLED.or(flags)

				// Destroy all proxies.
				val broadPhase = world.contactManager.broadPhase

				for (fixture in fixtureIterator) {
					(fixture as Fixture).destroyProxies(broadPhase)
				}

				// Destroy the attached contacts.
				for (edge in contactEdgeIterator) {
					world.contactManager.destroy(edge.contact)
				}

				contactEdge = null
			}
		}

	/**
	 * Set this body to have fixed rotation. This causes the mass to be reset.
	 */
	var isFixedRotation: Boolean
		get() = BodyFlags.FIXED_ROTATION.isit(flags)
		set(value) {
			if (value == isFixedRotation)
				return

			flags = BodyFlags.FIXED_ROTATION.update(flags, value)

			angularVelocity = 0.0
			resetMassData()
		}

	/**
	 * You can disable sleeping on this body. If you disable sleeping, the
	 * body will be woken.
	 */
	var allowAutoSleep: Boolean
		get() = BodyFlags.AUTO_SLEEP.isit(flags)
		set(value) {
			flags = BodyFlags.AUTO_SLEEP.update(flags, value)

			if (!value) {
				isAwake = true
			}
		}

	/**
	 * Set the type of this body. This may alter the mass and velocity.
	 */
	var type: BodyType = def.type
		set(value) {
			if (world.isLocked)
				throw ConcurrentModificationException()

			field = value

			resetMassData()

			if (value == BodyType.STATIC) {
				linearVelocity = Vector2d.ZERO
				angularVelocity = 0.0
				sweep.a0 = sweep.a
				sweep.c0 = sweep.c
				flags = BodyFlags.AWAKE.not(flags)
				synchronizeFixtures()
			}

			isAwake = true

			force = Vector2d.ZERO
			torque = 0.0

			var edge = contactEdge

			while (edge != null) {
				// TODO: проверить, что делает destroy
				world.contactManager.destroy(edge.contact)
				edge = edge.next
			}

			contactEdge = null

			val broadPhase = world.contactManager.broadPhase
			var f: Fixture? = fixtureList

			while (f != null) {
				for (proxy in f.proxies) {
					broadPhase.touchProxy(proxy.proxyId)
				}

				f = f.next
			}
		}

	/** Creates a fixture and attach it to this body. Use this function if you need
	 * to set some fixture parameters, like friction. Otherwise you can create the
	 * fixture directly from a shape.
	 * If the density is non-zero, this function automatically updates the mass of the body.
	 * Contacts are not created until the next time step.
	 * @param def the fixture definition.
	 * @warning This function is locked during callbacks.
	 */
	fun createFixture(def: FixtureDef): Fixture {
		if (world.isLocked)
			throw ConcurrentModificationException()

		val fixture = Fixture(this, def)

		if (isEnabled) {
			fixture.createProxies(world.contactManager.broadPhase, transform)
		}

		fixture.next = fixtureList
		fixtureList = fixture
		fixtureCount++

		// Adjust mass properties if needed.
		if (fixture.density > 0.0) {
			resetMassData()
		}

		// Let the world know we have a new fixture. This will cause new contacts
		// to be created at the beginning of the next time step.
		world.notifyNewContacts()

		return fixture
	}

	/**
	 * Creates a fixture from a shape and attach it to this body.
	 * This is a convenience function. Use b2FixtureDef if you need to set parameters
	 * like friction, restitution, user data, or filtering.
	 * If the density is non-zero, this function automatically updates the mass of the body.
	 * @param shape the shape to be cloned.
	 * @param density the shape density (set to zero for static bodies).
	 * @warning This function is locked during callbacks.
	 */
	fun createFixture(shape: IShape<*>, density: Double): Fixture {
		return createFixture(
			FixtureDef(
				shape = shape,
				density = density
			)
		)
	}

	/**
	 * Destroy a fixture. This removes the fixture from the broad-phase and
	 * destroys all contacts associated with this fixture. This will
	 * automatically adjust the mass of the body if the body is dynamic and the
	 * fixture has positive density.
	 * All fixtures attached to a body are implicitly destroyed when the body is destroyed.
	 * @param fixture the fixture to be removed.
	 * @warning This function is locked during callbacks.
	 */
	fun destroyFixture(fixture: Fixture) {
		if (world.isLocked)
			throw ConcurrentModificationException()

		require(fixture.body == this) { "$fixture does not belong to $this (belongs to ${fixture.body})" }
		check(fixtureCount > 0) { "Having no tracked fixtures, but $fixture belongs to us" }

		var node: Fixture? = fixtureList
		var found = false
		var previous: Fixture? = null

		while (node != null) {
			if (node == fixture) {
				// TODO: Это должно работать
				previous?.next = node.next
				found = true
				break
			} else {
				previous = node
				node = node.next
			}
		}

		check(found) { "Can't find $fixture in linked list of fixtures" }

		val density = node!!.density

		// Destroy any contacts associated with the fixture.
		var edge = contactEdge

		while (edge != null) {
			val contact = edge.contact
			edge = edge.next

			val fixtureA = contact.fixtureA
			val fixtureB = contact.fixtureB

			if (fixtureA == fixture || fixtureB == fixture) {
				// This destroys the contact and removes it from
				// this body's contact list.
				world.contactManager.destroy(contact)
			}
		}

		fixture as Fixture

		if (isEnabled) {
			fixture.destroyProxies(world.contactManager.broadPhase)
		}

		fixture.unlink()
		fixtureCount--

		// Reset the mass data
		if (density > 0.0) {
			resetMassData()
		}
	}

	/**
	 * This resets the mass properties to the sum of the mass properties of the fixtures.
	 * This normally does not need to be called unless you called SetMassData to override
	 * the mass and you later want to reset the mass.
	 */
	fun resetMassData() {
		// Compute mass data from shapes. Each shape has its own density.
		mass = 0.0
		invMass = 0.0
		rotInertia = 0.0
		rotInertiaInv = 0.0
		sweep.localCenter = Vector2d.ZERO

		if (type == BodyType.STATIC || type == BodyType.KINEMATIC) {
			sweep.c0 = transform.position
			sweep.c = transform.position
			sweep.a0 = sweep.a
			return
		}

		check(type == BodyType.DYNAMIC)

		// Accumulate mass over all fixtures.
		var localCenter = Vector2d.ZERO

		for (fixture in fixtureIterator) {
			if (fixture.density == 0.0) {
				continue
			}

			val massData = fixture.getMassData()
			mass += massData.mass
			localCenter += massData.center * massData.mass
			rotInertia += massData.inertia
		}

		// Compute center of mass.
		if (mass > 0.0) {
			invMass = 1.0 / mass
			localCenter *= invMass
		}

		if (rotInertia > 0.0 && !isFixedRotation) {
			// Center the inertia about the center of mass.
			rotInertia -= mass * localCenter.dot(localCenter)
			check(rotInertia > 0.0)
			rotInertiaInv = 1.0 / rotInertia
		} else {
			rotInertia = 0.0
			rotInertiaInv = 0.0
		}

		// Move center of mass.
		val oldCenter = sweep.c
		sweep.localCenter = localCenter
		sweep.c = transform.times(localCenter)
		sweep.c0 = sweep.c

		// Update center of mass velocity.
		linearVelocity += b2Cross(angularVelocity, sweep.c - oldCenter)
	}

	/**
	 * Get the mass data of the body.
	 * @return a struct containing the mass, inertia and center of the body.
	 * Set the mass properties to override the mass properties of the fixtures.
	 * Note that this changes the center of mass position.
	 * Note that creating or destroying fixtures can also alter the mass.
	 * This function has no effect if the body isn't dynamic.
	 * @param data the mass properties.
	 */
	var massData: MassData
		get() = MassData(
			mass = mass,
			inertia = inertia,
			center = sweep.localCenter
		)
		set(value) {
			if (world.isLocked)
				throw ConcurrentModificationException()

			if (type != BodyType.DYNAMIC)
				return

			invMass = 0.0
			rotInertia = 0.0
			rotInertiaInv = 0.0

			mass = value.mass

			if (mass <= 0.0) {
				mass = 1.0
			}

			invMass = 1.0 / mass

			if (value.inertia > 0.0 && !isFixedRotation) {
				rotInertia = value.inertia - mass * value.center.dot(value.center)
				check(rotInertia > 0.0)
				rotInertiaInv = 1.0 / rotInertia
			}

			// Move center of mass.
			val oldCenter = sweep.c
			sweep.localCenter = value.center
			sweep.c = transform.times(sweep.localCenter)
			sweep.c0 = sweep.c

			// Update center of mass velocity.
			linearVelocity += b2Cross(angularVelocity, sweep.c - oldCenter)
		}

	internal fun shouldCollide(other: Body): Boolean {
		// At least one body should be dynamic.
		if (type != BodyType.DYNAMIC && other.type != BodyType.DYNAMIC)
			return false

		// Does a joint prevent collision?
		var joint = jointList

		while (joint != null) {
			if (joint.otherNullable == other && !joint.joint.collideConnected) {
				return false
			}

			joint = joint.next
		}

		return true
	}

	init {
		sweep.c0 = transform.position
		sweep.c = transform.position
		sweep.a0 = def.angle
		sweep.a = def.angle
	}

	/**
	 * Set the position of the body's origin and rotation.
	 * Manipulating a body's transform may cause non-physical behavior.
	 * Note: contacts are updated on the next call to b2World::Step.
	 * @param position the world position of the body's local origin.
	 * @param angle the world rotation in radians.
	 */
	fun setTransform(position: Vector2d, angle: Double) {
		if (world.isLocked)
			throw ConcurrentModificationException()

		transform.rotation.set(angle)
		transform.position = position

		sweep.c = transform.times(sweep.localCenter)
		sweep.a = angle

		sweep.c0 = sweep.c
		sweep.a0 = angle

		val broadPhase = world.contactManager.broadPhase

		for (fixture in fixtureIterator) {
			(fixture as Fixture?)?.synchronize(broadPhase, transform, transform)
		}

		// Check for new contacts the next step
		world.notifyNewContacts()
	}

	/**
	 * Get the rotational inertia of the body about the local origin.
	 * @return the rotational inertia, usually in kg-m^2.
	 */
	val inertia: Double
		get() = rotInertia + mass * sweep.localCenter.dot(sweep.localCenter)

	/**
	 * Get the local position of the center of mass.
	 */
	val localCenter: Vector2d
		get() = sweep.localCenter

	/**
	 * Get the world position of the center of mass.
	 */
	val worldCenter: Vector2d
		get() = sweep.c

	/**
	 * Get the world coordinates of a point given the local coordinates.
	 * @param localPoint a point on the body measured relative the the body's origin.
	 * @return the same point expressed in world coordinates.
	 */
	fun getWorldPoint(localPoint: Vector2d): Vector2d {
		return transform.times(localPoint)
	}

	/**
	 * Get the world coordinates of a vector given the local coordinates.
	 * @param localVector a vector fixed in the body.
	 * @return the same vector expressed in world coordinates.
	 */
	fun getWorldVector(localPoint: Vector2d): Vector2d {
		return transform.rotation.times(localPoint)
	}

	/**
	 * Gets a local point relative to the body's origin given a world point.
	 * @param worldPoint a point in world coordinates.
	 * @return the corresponding local point relative to the body's origin.
	 */
	fun getLocalPoint(worldPoint: Vector2d): Vector2d {
		return transform.timesT(worldPoint)
	}

	/**
	 * Gets a local vector given a world vector.
	 * @param worldVector a vector in world coordinates.
	 * @return the corresponding local vector.
	 */
	fun getLocalVector(worldVector: Vector2d): Vector2d {
		return transform.rotation.timesT(worldVector)
	}

	/**
	 * Get the world linear velocity of a world point attached to this body.
	 * @param worldPoint a point in world coordinates.
	 * @return the world velocity of a point.
	 */
	fun getLinearVelocityFromWorldPoint(worldPoint: Vector2d): Vector2d {
		return linearVelocity + b2Cross(angularVelocity, worldPoint - sweep.c)
	}

	/**
	 * Get the world velocity of a local point.
	 * @param localPoint a point in local coordinates.
	 * @return the world velocity of a point.
	 */
	fun getLinearVelocityFromLocalPoint(localPoint: Vector2d): Vector2d {
		return getLinearVelocityFromWorldPoint(getWorldPoint(localPoint))
	}

	/**
	 * Apply a force at a world point. If the force is not
	 * applied at the center of mass, it will generate a torque and
	 * affect the angular velocity. This wakes up the body.
	 * @param force the world force vector, usually in Newtons (N).
	 * @param point the world position of the point of application.
	 * @param wake also wake up the body
	 */
	fun applyForce(force: Vector2d, point: Vector2d, wake: Boolean) {
		if (type != BodyType.DYNAMIC)
			return

		if (wake && !isAwake)
			isAwake = true

		// Don't accumulate a force if the body is sleeping.
		if (isAwake) {
			this.force += force
			this.torque += b2Cross(point - sweep.c, force)
		}
	}

	/**
	 * Apply a force to the center of mass. This wakes up the body.
	 * @param force the world force vector, usually in Newtons (N).
	 * @param wake also wake up the body
	 */
	fun applyForceToCenter(force: Vector2d, wake: Boolean) {
		if (type != BodyType.DYNAMIC)
			return

		if (wake && !isAwake)
			isAwake = true

		// Don't accumulate a force if the body is sleeping.
		if (isAwake) {
			this.force += force
		}
	}

	/**
	 * Apply a torque. This affects the angular velocity
	 * without affecting the linear velocity of the center of mass.
	 * @param torque about the z-axis (out of the screen), usually in N-m.
	 * @param wake also wake up the body
	 */
	fun applyTorque(torque: Double, wake: Boolean) {
		if (type != BodyType.DYNAMIC)
			return

		if (wake && !isAwake)
			isAwake = true

		if (isAwake) {
			this.torque += torque
		}
	}

	/**
	 * Apply an impulse at a point. This immediately modifies the velocity.
	 * It also modifies the angular velocity if the point of application
	 * is not at the center of mass. This wakes up the body.
	 * @param impulse the world impulse vector, usually in N-seconds or kg-m/s.
	 * @param point the world position of the point of application.
	 * @param wake also wake up the body
	 */
	fun applyLinearImpulse(impulse: Vector2d, point: Vector2d, wake: Boolean) {
		if (type != BodyType.DYNAMIC)
			return

		if (wake && !isAwake)
			isAwake = true

		if (isAwake) {
			linearVelocity += impulse * invMass
			angularVelocity += rotInertiaInv *  b2Cross(point - sweep.c, impulse)
		}
	}

	/**
	 * Apply an impulse to the center of mass. This immediately modifies the velocity.
	 * @param impulse the world impulse vector, usually in N-seconds or kg-m/s.
	 * @param wake also wake up the body
	 */
	fun applyLinearImpulseToCenter(impulse: Vector2d, wake: Boolean) {
		if (type != BodyType.DYNAMIC)
			return

		if (wake && !isAwake)
			isAwake = true

		if (isAwake) {
			linearVelocity += impulse * invMass
		}
	}

	/**
	 * Apply an angular impulse.
	 * @param impulse the angular impulse in units of kg*m*m/s
	 * @param wake also wake up the body
	 */
	fun applyAngularImpulse(impulse: Double, wake: Boolean) {
		if (type != BodyType.DYNAMIC)
			return

		if (wake && !isAwake)
			isAwake = true

		if (isAwake) {
			angularVelocity += impulse * rotInertiaInv
		}
	}

	internal fun synchronizeTransform() {
		transform.rotation.set(sweep.a)
		transform.position = sweep.c - transform.rotation.times(sweep.localCenter)
	}

	internal fun advance(alpha: Double) {
		sweep.advance(alpha)
		sweep.c = sweep.c0
		sweep.a = sweep.a0
		synchronizeTransform()
	}

	internal fun synchronizeFixtures() {
		val broadPhase = world.contactManager.broadPhase

		if (isAwake) {
			val transform1 = Transform()
			transform1.rotation.set(sweep.a0)
			transform1.position = sweep.c0 - transform1.rotation.times(sweep.localCenter)

			for (fixture in fixtureIterator) {
				(fixture as Fixture).synchronize(broadPhase, transform1, transform)
			}
		} else {
			for (fixture in fixtureIterator) {
				(fixture as Fixture).synchronize(broadPhase, transform, transform)
			}
		}
	}

	fun dump() {
		TODO("Not yet implemented")
	}
}
