package ru.dbotthepony.kbox2d.dynamics

import ru.dbotthepony.kbox2d.api.*
import ru.dbotthepony.kbox2d.collision.BroadPhase
import ru.dbotthepony.kbox2d.dynamics.contact.AbstractContact

/**
 * Delegate of b2World.
 */
class ContactManager {
	val broadPhase = BroadPhase()
	var contactList: AbstractContact? = null
	var contactCount: Int = 0
		private set

	var contactFilter: IContactFilter? = null
	var contactListener: IContactListener? = null

	val contactListIterator: Iterator<AbstractContact> get() {
		return object : Iterator<AbstractContact> {
			private var node = contactList

			override fun hasNext(): Boolean {
				return node != null
			}

			override fun next(): AbstractContact {
				val old = node!!
				node = old.next
				return old
			}
		}
	}

	fun destroy(contact: AbstractContact) {
		val fixtureA = contact.fixtureA
		val fixtureB = contact.fixtureB
		val bodyA = fixtureA.body as Body
		val bodyB = fixtureB.body as Body

		if (contact.isTouching) {
			contactListener?.endContact(contact)
		}

		// Remove from the world.
		run {
			val prev = contact.prev as AbstractContact?
			val next = contact.next as AbstractContact?

			prev?.next = next
			next?.prev = prev

			if (contactList == contact) {
				contactList = next ?: prev
			}
		}

		// Remove from body 1
		run {
			val prev = contact.nodeA.prev
			val next = contact.nodeA.next

			prev?.next = next
			next?.prev = prev

			if (contact.nodeA == bodyA.contactEdge) {
				bodyA.contactEdge = next ?: prev
			}
		}

		// Remove from body 2
		run {
			val prev = contact.nodeB.prev
			val next = contact.nodeB.next

			prev?.next = next
			next?.prev = prev

			if (contact.nodeB == bodyB.contactEdge) {
				bodyB.contactEdge = next ?: prev
			}
		}

		contactCount--
	}

	/**
	 * This is the top level collision call for the time step. Here
	 * all the narrow phase collision is processed for the world
	 * contact list.
	 */
	fun collide() {
		// Update awake contacts.

		for (c in contactListIterator) {
			c as AbstractContact
			val fixtureA = c.fixtureA
			val fixtureB = c.fixtureB
			val indexA = c.childIndexA
			val indexB = c.childIndexB
			val bodyA = fixtureA.body as Body
			val bodyB = fixtureB.body as Body

			// Is this contact flagged for filtering?
			if (c.isFlaggedForFiltering) {
				// Should these bodies collide?
				if (!bodyB.shouldCollide(bodyA)) {
					destroy(c)
					continue
				}

				// Check user filtering.
				if (contactFilter?.shouldCollide(fixtureA, fixtureB) == false) {
					destroy(c)
					continue
				}

				// Clear the filtering flag.
				c.isFlaggedForFiltering = false
			}

			val activeA = bodyA.isAwake && bodyA.type != ru.dbotthepony.kbox2d.api.BodyType.STATIC
			val activeB = bodyB.isAwake && bodyB.type != ru.dbotthepony.kbox2d.api.BodyType.STATIC

			// At least one body must be awake and it must be dynamic or kinematic.
			if (!activeA && !activeB) {
				continue
			}

			val proxyIdA = fixtureA.proxies[indexA].proxyId
			val proxyIdB = fixtureB.proxies[indexB].proxyId
			val overlap = broadPhase.testOverlap(proxyIdA, proxyIdB)

			if (!overlap) {
				destroy(c)
				continue
			}

			c.update(contactListener)
		}
	}

	fun findNewContacts() {
		broadPhase.updatePairs(this::addPair)
	}

	fun addPair(proxyUserDataA: Any?, proxyUserDataB: Any?) {
		val proxyA = proxyUserDataA as FixtureProxy
		val proxyB = proxyUserDataB as FixtureProxy

		val fixtureA = proxyA.fixture
		val fixtureB = proxyB.fixture

		val indexA = proxyA.childIndex
		val indexB = proxyB.childIndex

		val bodyA = fixtureA.body as Body
		val bodyB = fixtureB.body as Body

		// Are the fixtures on the same body?
		if (bodyA === bodyB) {
			return
		}

		// TODO_ERIN use a hash table to remove a potential bottleneck when both
		// bodies have a lot of contacts.
		// Does a contact already exist?
		for (edge in bodyB.contactEdgeIterator) {
			if (edge.other === bodyA) {
				val fA = edge.contact.fixtureA
				val fB = edge.contact.fixtureB
				val iA = edge.contact.childIndexA
				val iB = edge.contact.childIndexB

				if (fA === fixtureA && fB === fixtureB && iA == indexA && iB == indexB) {
					// A contact already exists.
					return
				}

				if (fA === fixtureB && fB === fixtureA && iA == indexB && iB == indexA) {
					// A contact already exists.
					return
				}
			}
		}

		// Does a joint override collision? Is at least one body dynamic?
		if (!bodyB.shouldCollide(bodyA)) {
			return
		}

		// Check user filtering.
		if (contactFilter?.shouldCollide(fixtureA, fixtureB) == false) {
			return
		}

		// Call the factory.
		val c = AbstractContact.create(fixtureA, indexA, fixtureB, indexB)

		// Contact creation may swap fixtures.
		// Insert into the world.
		c.next = contactList
		(contactList as AbstractContact?)?.prev = c
		contactList = c

		// Connect to island graph.
		// (connection is done in AbstractContact initializer)
		contactCount++
	}
}