
@file:Suppress("nothing_to_inline", "unused")

package ru.dbotthepony.kvector.api

/**
 * Classes implementing this interface can be manipulated like matrices
 */
interface IMatrixLike {
	/**
	 * Rows this matrix has
	 */
	val rows: Int

	/**
	 * Columns this matrix has
	 */
	val columns: Int

	/**
	 * Whenever is this matrix a square matrix
	 */
	val isSquareMatrix: Boolean get() = rows == columns

	/**
	 * @return Whenever [other] has the same dimensions as this matrix
	 */
	fun sizeEquals(other: IMatrixLike): Boolean {
		return rows == other.rows && columns == other.columns
	}

	/**
	 * @throws IllegalArgumentException if [other] is not the same dimensions as this matrix
	 */
	fun requireSizeEquals(other: IMatrixLike) {
		if (!sizeEquals(other)) {
			throw IllegalArgumentException("Other matrix (${other.columns}x${other.rows}) is not the same dimensions as this matrix (${columns}x${rows})")
		}
	}

	/**
	 * @throws IllegalArgumentException if [other] is not the same dimensions as this matrix
	 */
	fun requireSizeEquals(other: IMatrixLike, lazy: () -> Any) {
		if (!sizeEquals(other)) {
			throw IllegalArgumentException(lazy.invoke().toString())
		}
	}

	/**
	 * @throws IllegalStateException if [other] is not the same dimensions as this matrix
	 */
	fun checkSizeEquals(other: IMatrixLike) {
		if (!sizeEquals(other)) {
			throw IllegalStateException("Other matrix (${other.columns}x${other.rows}) is not the same dimensions as this matrix (${columns}x${rows})")
		}
	}

	/**
	 * @throws IllegalStateException if [other] is not the same dimensions as this matrix
	 */
	fun checkSizeEquals(other: IMatrixLike, lazy: () -> Any) {
		if (!sizeEquals(other)) {
			throw IllegalStateException(lazy.invoke().toString())
		}
	}

	/**
	 * @return Whenever [other] has the same amount of rows as this matrix
	 */
	fun rowsEquals(other: IMatrixLike): Boolean {
		return rows == other.rows
	}

	/**
	 * @throws IllegalArgumentException if [other] is not the same amount of rows as this matrix
	 */
	fun requireRowsEquals(other: IMatrixLike) {
		if (!rowsEquals(other)) {
			throw IllegalArgumentException("Other matrix (${other.rows}) has different amount of rows compared to this matrix (${rows})")
		}
	}

	/**
	 * @throws IllegalArgumentException if [other] is not the same amount of rows as this matrix
	 */
	fun requireRowsEquals(other: IMatrixLike, lazy: () -> Any) {
		if (!rowsEquals(other)) {
			throw IllegalArgumentException(lazy.invoke().toString())
		}
	}

	/**
	 * @throws IllegalStateException if [other] is not the same amount of rows as this matrix
	 */
	fun checkRowsEquals(other: IMatrixLike) {
		if (!rowsEquals(other)) {
			throw IllegalStateException("Other matrix (${other.rows}) has different amount of rows compared to this matrix (${rows})")
		}
	}

	/**
	 * @throws IllegalStateException if [other] is not the same amount of rows as this matrix
	 */
	fun checkRowsEquals(other: IMatrixLike, lazy: () -> Any) {
		if (!rowsEquals(other)) {
			throw IllegalStateException(lazy.invoke().toString())
		}
	}

	/**
	 * @return Whenever [other] has the same amount of columns as this matrix
	 */
	fun columnsEquals(other: IMatrixLike): Boolean {
		return columns == other.columns
	}

	/**
	 * @throws IllegalArgumentException if [other] is not the same amount of columns as this matrix
	 */
	fun requireColumnsEquals(other: IMatrixLike) {
		if (!rowsEquals(other)) {
			throw IllegalArgumentException("Other matrix (${other.columns}) has different amount of columns compared to this matrix (${columns})")
		}
	}

	/**
	 * @throws IllegalArgumentException if [other] is not the same amount of columns as this matrix
	 */
	fun requireColumnsEquals(other: IMatrixLike, lazy: () -> Any) {
		if (!rowsEquals(other)) {
			throw IllegalArgumentException(lazy.invoke().toString())
		}
	}

	/**
	 * @throws IllegalStateException if [other] is not the same amount of columns as this matrix
	 */
	fun checkColumnsEquals(other: IMatrixLike) {
		if (!rowsEquals(other)) {
			throw IllegalStateException("Other matrix (${other.columns}) has different amount of columns compared to this matrix (${columns})")
		}
	}

	/**
	 * @throws IllegalStateException if [other] is not the same amount of columns as this matrix
	 */
	fun checkColumnsEquals(other: IMatrixLike, lazy: () -> Any) {
		if (!rowsEquals(other)) {
			throw IllegalStateException(lazy.invoke().toString())
		}
	}
}

interface IMatrixGetterInt : IMatrixLike {
	/**
	 * @return component of this matrix, as int
	 */
	operator fun get(column: Int, row: Int): Int
}

interface IMatrixGetterLong : IMatrixLike {
	/**
	 * @return component of this matrix, as long
	 */
	operator fun get(column: Int, row: Int): Long
}

interface IMatrixGetterShort : IMatrixLike {
	/**
	 * @return component of this matrix, as short
	 */
	operator fun get(column: Int, row: Int): Short
}

interface IMatrixGetterByte : IMatrixLike {
	/**
	 * @return component of this matrix, as byte
	 */
	operator fun get(column: Int, row: Int): Byte
}

interface IMatrixSetterInt : IMatrixLike {
	/**
	 * Sets component of this matrix, as int
	 */
	operator fun set(column: Int, row: Int, value: Int)
}

interface IMatrixSetterLong : IMatrixLike {
	/**
	 * Sets component of this matrix, as long
	 */
	operator fun set(column: Int, row: Int, value: Long)
}

interface IMatrixSetterShort : IMatrixLike {
	/**
	 * Sets component of this matrix, as short
	 */
	operator fun set(column: Int, row: Int, value: Short)
}

interface IMatrixSetterByte : IMatrixLike {
	/**
	 * Sets component of this matrix, as byte
	 */
	operator fun set(column: Int, row: Int, value: Byte)
}

/**
 * Root inheritance tree of Matrix classes
 */
interface IMatrix<T : IMatrix<T>> : IMatrixLike {
	/**
	 * Whenever all components of this matrix are finite
	 */
	val isFinite: Boolean

	/**
	 * Whenever any of components of this matrix are NaN
	 */
	val isNaN: Boolean

	/**
	 * Whenever is this matrix valid (components are not NaN and are finite)
	 */
	val isValid: Boolean get() = isFinite && !isNaN

	/**
	 * @throws IllegalArgumentException if matrix is not finite
	 */
	fun requireIsFinite(lambda: () -> Any) {
		if (!isFinite) {
			throw IllegalArgumentException(lambda.invoke().toString())
		}
	}

	/**
	 * @throws IllegalArgumentException if matrix is not finite
	 */
	fun requireIsFinite() {
		if (!isFinite) {
			throw IllegalArgumentException("Matrix is not finite")
		}
	}

	/**
	 * @throws IllegalStateException if matrix is not finite
	 */
	fun checkIsFinite(lambda: () -> Any) {
		if (!isFinite) {
			throw IllegalStateException(lambda.invoke().toString())
		}
	}

	/**
	 * @throws IllegalStateException if matrix is not finite
	 */
	fun checkIsFinite() {
		if (!isFinite) {
			throw IllegalStateException("Matrix is not finite")
		}
	}

	/**
	 * @throws IllegalArgumentException if vector has NaN component
	 */
	fun requireNotNaN(lambda: () -> Any) {
		if (isNaN) {
			throw IllegalArgumentException(lambda.invoke().toString())
		}
	}

	/**
	 * @throws IllegalArgumentException if matrix has NaN component
	 */
	fun requireNotNaN() {
		if (isNaN) {
			throw IllegalArgumentException("Matrix has NaN component")
		}
	}

	/**
	 * @throws IllegalStateException if matrix has NaN component
	 */
	fun checkNotNaN(lambda: () -> Any) {
		if (!isFinite) {
			throw IllegalStateException(lambda.invoke().toString())
		}
	}

	/**
	 * @throws IllegalStateException if matrix has NaN component
	 */
	fun checkNotNaN() {
		if (!isFinite) {
			throw IllegalStateException("Matrix has NaN component")
		}
	}

	/**
	 * @throws IllegalArgumentException if matrix has NaN or infinite component
	 */
	fun requireIsValid() {
		requireIsFinite()
		requireNotNaN()
	}

	/**
	 * @throws IllegalStateException if matrix has NaN or infinite component
	 */
	fun checkIsValid() {
		requireIsFinite()
		requireNotNaN()
	}

	/**
	 * @throws IllegalArgumentException if matrix has NaN or infinite component
	 */
	fun requireIsValid(lambda: () -> Any) {
		requireIsFinite(lambda)
		requireNotNaN(lambda)
	}

	/**
	 * @throws IllegalStateException if matrix has NaN or infinite component
	 */
	fun checkIsValid(lambda: () -> Any) {
		requireIsFinite(lambda)
		requireNotNaN(lambda)
	}
}


/**
 * Marks matrix for being transposable with result of T
 */
interface ITransposable<T> {
	/**
	 * New matrix, which is transposed variant of this matrix
	 *
	 * @return new matrix
	 */
	val transposed: T
}

/**
 * Marks matrix for being transposable in place
 */
interface IMutableTransposable<T> {
	/**
	 * Attempts to transpose this matrix, writing the transpose result into itself
	 *
	 * @throws IllegalStateException if this matrix is not a square matrix
	 * @return this matrix
	 */
	fun transpose(): T
}

/**
 * Getter for matrix complement
 */
interface IMatrixComplement<T> {
	/**
	 * Constructs a new matrix, representing complement of this matrix at [column] and [row]
	 *
	 * @throws IllegalStateException if this matrix is not applicable for complement operation
	 * @return new matrix, representing minor, with size - 1 of this matrix
	 */
	fun complement(column: Int, row: Int): T
}

/**
 * Defines common operations with concrete square matrices, such as [translation], [scale]ing, etc
 */
interface ISquareMatrix<T : ISquareMatrix<T, V, F>, V, F> {
	/**
	 * Current translation of this matrix.
	 */
	val translation: V

	/**
	 * Does raw translation of this matrix by specified [vector],
	 * returning new matrix as result.
	 *
	 * If you scaled and/or rotated this matrix, and want to move it
	 * with scale and translation, use [translateWithMultiplication].
	 *
	 * @return new matrix
	 */
	fun translate(vector: V): T

	/**
	 * Does translation of this matrix by specified [vector],
	 * with accounting of it's state, returning new matrix as result.
	 *
	 * @return new matrix
	 */
	fun translateWithMultiplication(vector: V): T

	/**
	 * Multiplies main diagonal of this matrix with [vector],
	 * returning new matrix as result.
	 *
	 * @return new matrix
	 */
	fun scale(vector: F): T
}

/**
 * Defines common mutable operations with concrete square matrices, such as [translation], [scale]ing, etc
 */
interface IMutableSquareMatrix<T : IMutableSquareMatrix<T, V, F>, V, F> : ISquareMatrix<T, V, F> {
	/**
	 * Current translation of this matrix.
	 *
	 * Writing vector to this property will set raw matrix translation.
	 */
	override var translation: V

	/**
	 * Does raw translation of this matrix by specified [vector],
	 * writing result into this matrix.
	 *
	 * If you scaled and/or rotated this matrix, and want to move it
	 * with scale and translation, use [translateWithMultiplication].
	 *
	 * @return this matrix
	 */
	fun translateMut(vector: V): T

	/**
	 * Does translation of this matrix by specified [vector],
	 * with accounting of it's state, writing result into this matrix.
	 *
	 * @return this matrix
	 */
	fun translateWithMultiplicationMut(vector: V): T

	/**
	 * Multiplies main diagonal of this matrix with [vector],
	 * writing result into this matrix.
	 *
	 * @return this matrix
	 */
	fun scaleMut(vector: F): T
}
