
@file:Suppress("unused")

package ru.dbotthepony.kvector.api.concrete

import ru.dbotthepony.kvector.api.IFloatMatrix
import ru.dbotthepony.kvector.api.IMatrix
import ru.dbotthepony.kvector.api.ISquareMatrix
import ru.dbotthepony.kvector.api.ITransposable
import ru.dbotthepony.kvector.matrix.*
import ru.dbotthepony.kvector.matrix.ndouble.*
import ru.dbotthepony.kvector.matrix.nfloat.*
import ru.dbotthepony.kvector.vector.nfloat.*
import java.nio.ByteBuffer
import java.nio.FloatBuffer

/**
 * [Matrix2f] and [MutableMatrix2f] implement this
 */
interface IMatrix2f<T : IMatrix2f<T>> : IMatrix<T>, IFloatMatrix<T>, ITransposable<T> {
	val c00: Float
	val c10: Float
	val c01: Float
	val c11: Float
	val r00: Float
	val r01: Float
	val r10: Float
	val r11: Float
	val a11: Float
	val a21: Float
	val a12: Float
	val a22: Float
	val b11: Float
	val b21: Float
	val b12: Float
	val b22: Float
	val c0: Vector2f
	val c1: Vector2f
	val r0: Vector2f
	val r1: Vector2f

	/**
	 * Creates a copy of this matrix as [Matrix4d],
	 * setting last two rows and columns to 0,
	 * casting all [Float]s to [Double]s
	 */
	fun toMatrix4d(): Matrix4d {
		return Matrix4d(
			c00.toDouble(), c01.toDouble(), 0.0, 0.0,
			c10.toDouble(), c11.toDouble(), 0.0, 0.0,
			0.0, 0.0, 0.0, 0.0,
			0.0, 0.0, 0.0, 0.0,
		)
	}

	/**
	 * Creates a copy of this matrix as [Matrix3d],
	 * setting last row and column to 0,
	 * casting all [Float]s to [Double]s
	 */
	fun toMatrix3d(): Matrix3d {
		return Matrix3d(
			c00.toDouble(), c01.toDouble(), 0.0,
			c10.toDouble(), c11.toDouble(), 0.0,
			0.0, 0.0, 0.0,
		)
	}

	/**
	 * Creates a copy of this matrix as [Matrix2d],
	 * casting all [Float]s to [Double]s
	 */
	fun toMatrix2d(): Matrix2d {
		return Matrix2d(
			c00.toDouble(), c01.toDouble(),
			c10.toDouble(), c11.toDouble(),
		)
	}

	/**
	 * Creates a copy of this matrix as [MutableMatrix4d],
	 * setting last two columns and rows to 0,
	 * casting all [Float]s to [Double]s
	 */
	fun toMutableMatrix4d(): MutableMatrix4d {
		return MutableMatrix4d(
			c00.toDouble(), c01.toDouble(), 0.0, 0.0,
			c10.toDouble(), c11.toDouble(), 0.0, 0.0,
			0.0, 0.0, 0.0, 0.0,
			0.0, 0.0, 0.0, 0.0,
		)
	}

	/**
	 * Creates a copy of this matrix as [MutableMatrix3d],
	 * setting last column and row to 0,
	 * casting all [Float]s to [Double]s
	 */
	fun toMutableMatrix3d(): MutableMatrix3d {
		return MutableMatrix3d(
			c00.toDouble(), c01.toDouble(), 0.0,
			c10.toDouble(), c11.toDouble(), 0.0,
			0.0, 0.0, 0.0,
		)
	}

	/**
	 * Creates a copy of this matrix as [MutableMatrix2d],
	 * casting all [Float]s to [Double]s
	 */
	fun toMutableMatrix2d(): MutableMatrix2d {
		return MutableMatrix2d(
			c00.toDouble(), c01.toDouble(),
			c10.toDouble(), c11.toDouble(),
		)
	}

	/**
	 * Creates a copy of this matrix as [Matrix4f],
	 * setting last two columns and rows to 0
	 */
	fun toMatrix4f(): Matrix4f {
		return Matrix4f(
			c00, c01, 0f, 0f,
			c10, c11, 0f, 0f,
			0f, 0f, 0f, 0f,
			0f, 0f, 0f, 0f,
		)
	}

	/**
	 * Creates a copy of this matrix as [Matrix3f],
	 * setting last column and row to 0
	 */
	fun toMatrix3f(): Matrix3f {
		return Matrix3f(
			c00, c01, 0f,
			c10, c11, 0f,
			0f, 0f, 0f,
		)
	}

	/**
	 * Creates a copy of this matrix as [Matrix2f],
	 * truncating last row and column
	 */
	fun toMatrix2f(): Matrix2f {
		return Matrix2f(
			c00, c01,
			c10, c11,
		)
	}

	/**
	 * Creates a copy of this matrix as [MutableMatrix4f],
	 * setting last two columns and rows to 0,
	 */
	fun toMutableMatrix4f(): MutableMatrix4f {
		return MutableMatrix4f(
			c00, c01, 0f, 0f,
			c10, c11, 0f, 0f,
			0f, 0f, 0f, 0f,
			0f, 0f, 0f, 0f,
		)
	}

	/**
	 * Creates a copy of this matrix as [MutableMatrix3f],
	 * setting last column and row to 0,
	 * casing all [Double]s to [Float]s
	 */
	fun toMutableMatrix3f(): MutableMatrix3f {
		return MutableMatrix3f(
			c00, c01, 0f,
			c10, c11, 0f,
			0f, 0f, 0f,
		)
	}

	/**
	 * Creates a copy of this matrix as [MutableMatrix2f],
	 * truncating last row and column,
	 * casing all [Double]s to [Float]s
	 */
	fun toMutableMatrix2f(): MutableMatrix2f {
		return MutableMatrix2f(
			c00, c01,
			c10, c11,
		)
	}
}

/**
 * [Matrix3f] and [MutableMatrix3f] implement this
 */
interface IMatrix3f<T : IMatrix3f<T>> : IMatrix<T>, IFloatMatrix<T>, ITransposable<T>, ISquareMatrix<T, Vector2f, Vector3f> {
	val c00: Float
	val c10: Float
	val c20: Float
	val c01: Float
	val c11: Float
	val c21: Float
	val c02: Float
	val c12: Float
	val c22: Float
	val r00: Float
	val r01: Float
	val r02: Float
	val r10: Float
	val r11: Float
	val r12: Float
	val r20: Float
	val r21: Float
	val r22: Float
	val a11: Float
	val a21: Float
	val a31: Float
	val a12: Float
	val a22: Float
	val a32: Float
	val a13: Float
	val a23: Float
	val a33: Float
	val b11: Float
	val b21: Float
	val b31: Float
	val b12: Float
	val b22: Float
	val b32: Float
	val b13: Float
	val b23: Float
	val b33: Float
	val c0: Vector3f
	val c1: Vector3f
	val c2: Vector3f
	val r0: Vector3f
	val r1: Vector3f
	val r2: Vector3f

	override val translation: Vector2f get() = Vector2f(r02, r12)

	/**
	 * Creates a copy of this matrix as [Matrix4d],
	 * setting last row and column to 0,
	 * casting all [Float]s to [Double]s
	 */
	fun toMatrix4d(): Matrix4d {
		return Matrix4d(
			c00.toDouble(), c01.toDouble(), c02.toDouble(), 0.0,
			c10.toDouble(), c11.toDouble(), c12.toDouble(), 0.0,
			c20.toDouble(), c21.toDouble(), c22.toDouble(), 0.0,
			0.0, 0.0, 0.0, 0.0,
		)
	}

	/**
	 * Creates a copy of this matrix as [Matrix3d],
	 * casting all [Float]s to [Double]s
	 */
	fun toMatrix3d(): Matrix3d {
		return Matrix3d(
			c00.toDouble(), c01.toDouble(), c02.toDouble(),
			c10.toDouble(), c11.toDouble(), c12.toDouble(),
			c20.toDouble(), c21.toDouble(), c22.toDouble(),
		)
	}

	/**
	 * Creates a copy of this matrix as [Matrix2d],
	 * truncating last row and column,
	 * casting all [Float]s to [Double]s
	 */
	fun toMatrix2d(): Matrix2d {
		return Matrix2d(
			c00.toDouble(), c01.toDouble(),
			c10.toDouble(), c11.toDouble(),
		)
	}

	/**
	 * Creates a copy of this matrix as [MutableMatrix4d],
	 * setting last column and row to 0,
	 * casting all [Float]s to [Double]s
	 */
	fun toMutableMatrix4d(): MutableMatrix4d {
		return MutableMatrix4d(
			c00.toDouble(), c01.toDouble(), c02.toDouble(), 0.0,
			c10.toDouble(), c11.toDouble(), c12.toDouble(), 0.0,
			c20.toDouble(), c21.toDouble(), c22.toDouble(), 0.0,
			0.0, 0.0, 0.0, 0.0,
		)
	}

	/**
	 * Creates a copy of this matrix as [MutableMatrix3d],
	 * casting all [Float]s to [Double]s
	 */
	fun toMutableMatrix3d(): MutableMatrix3d {
		return MutableMatrix3d(
			c00.toDouble(), c01.toDouble(), c02.toDouble(),
			c10.toDouble(), c11.toDouble(), c12.toDouble(),
			c20.toDouble(), c21.toDouble(), c22.toDouble(),
		)
	}

	/**
	 * Creates a copy of this matrix as [MutableMatrix2d],
	 * truncating last row and column,
	 * casting all [Float]s to [Double]s
	 */
	fun toMutableMatrix2d(): MutableMatrix2d {
		return MutableMatrix2d(
			c00.toDouble(), c01.toDouble(),
			c10.toDouble(), c11.toDouble(),
		)
	}

	/**
	 * Creates a copy of this matrix as [Matrix4f],
	 * setting last column and row to 0
	 */
	fun toMatrix4f(): Matrix4f {
		return Matrix4f(
			c00, c01, c02, 0f,
			c10, c11, c12, 0f,
			c20, c21, c22, 0f,
			0f, 0f, 0f, 0f,
		)
	}

	/**
	 * Creates a copy of this matrix as [Matrix3f]
	 */
	fun toMatrix3f(): Matrix3f {
		return Matrix3f(
			c00, c01, c02,
			c10, c11, c12,
			c20, c21, c22,
		)
	}

	/**
	 * Creates a copy of this matrix as [Matrix2f],
	 * truncating last row and column
	 */
	fun toMatrix2f(): Matrix2f {
		return Matrix2f(
			c00, c01,
			c10, c11,
		)
	}

	/**
	 * Creates a copy of this matrix as [MutableMatrix4f],
	 * setting last column and row to 0
	 */
	fun toMutableMatrix4f(): MutableMatrix4f {
		return MutableMatrix4f(
			c00, c01, c02, 0f,
			c10, c11, c12, 0f,
			c20, c21, c22, 0f,
			0f, 0f, 0f, 0f,
		)
	}

	/**
	 * Creates a copy of this matrix as [MutableMatrix3f]
	 */
	fun toMutableMatrix3f(): MutableMatrix3f {
		return MutableMatrix3f(
			c00, c01, c02,
			c10, c11, c12,
			c20, c21, c22,
		)
	}

	/**
	 * Creates a copy of this matrix as [MutableMatrix2f],
	 * truncating last row and column
	 */
	fun toMutableMatrix2f(): MutableMatrix2f {
		return MutableMatrix2f(
			c00, c01,
			c10, c11,
		)
	}
}

/**
 * [Matrix4f] and [MutableMatrix4f] implement this
 */
interface IMatrix4f<T : IMatrix4f<T>> : IMatrix<T>, IFloatMatrix<T>, ITransposable<T>, ISquareMatrix<T, Vector3f, Vector4f> {
	val c00: Float
	val c10: Float
	val c20: Float
	val c30: Float
	val c01: Float
	val c11: Float
	val c21: Float
	val c31: Float
	val c02: Float
	val c12: Float
	val c22: Float
	val c32: Float
	val c03: Float
	val c13: Float
	val c23: Float
	val c33: Float
	val r00: Float
	val r01: Float
	val r02: Float
	val r03: Float
	val r10: Float
	val r11: Float
	val r12: Float
	val r13: Float
	val r20: Float
	val r21: Float
	val r22: Float
	val r23: Float
	val r30: Float
	val r31: Float
	val r32: Float
	val r33: Float
	val a11: Float
	val a21: Float
	val a31: Float
	val a41: Float
	val a12: Float
	val a22: Float
	val a32: Float
	val a42: Float
	val a13: Float
	val a23: Float
	val a33: Float
	val a43: Float
	val a14: Float
	val a24: Float
	val a34: Float
	val a44: Float
	val b11: Float
	val b21: Float
	val b31: Float
	val b41: Float
	val b12: Float
	val b22: Float
	val b32: Float
	val b42: Float
	val b13: Float
	val b23: Float
	val b33: Float
	val b43: Float
	val b14: Float
	val b24: Float
	val b34: Float
	val b44: Float
	val c0: Vector4f
	val c1: Vector4f
	val c2: Vector4f
	val c3: Vector4f
	val r0: Vector4f
	val r1: Vector4f
	val r2: Vector4f
	val r3: Vector4f

	override val translation: Vector3f get() = Vector3f(r03, r13, r23)

	/**
	 * Creates a copy of this matrix as [Matrix4d],
	 * casting all [Float]s to [Double]s
	 */
	fun toMatrix4d(): Matrix4d {
		return Matrix4d(
			c00.toDouble(), c01.toDouble(), c02.toDouble(), c03.toDouble(),
			c10.toDouble(), c11.toDouble(), c12.toDouble(), c13.toDouble(),
			c20.toDouble(), c21.toDouble(), c22.toDouble(), c23.toDouble(),
			c30.toDouble(), c31.toDouble(), c32.toDouble(), c33.toDouble(),
		)
	}

	/**
	 * Creates a copy of this matrix as [Matrix3d],
	 * truncating last row and column,
	 * casting all [Float]s to [Double]s
	 */
	fun toMatrix3d(): Matrix3d {
		return Matrix3d(
			c00.toDouble(), c01.toDouble(), c02.toDouble(),
			c10.toDouble(), c11.toDouble(), c12.toDouble(),
			c20.toDouble(), c21.toDouble(), c22.toDouble(),
		)
	}

	/**
	 * Creates a copy of this matrix as [Matrix2d],
	 * truncating two last rows and columns,
	 * casting all [Float]s to [Double]s
	 */
	fun toMatrix2d(): Matrix2d {
		return Matrix2d(
			c00.toDouble(), c01.toDouble(),
			c10.toDouble(), c11.toDouble(),
		)
	}

	/**
	 * Creates a copy of this matrix as [MutableMatrix4d],
	 * casting all [Float]s to [Double]s
	 */
	fun toMutableMatrix4d(): MutableMatrix4d {
		return MutableMatrix4d(
			c00.toDouble(), c01.toDouble(), c02.toDouble(), c03.toDouble(),
			c10.toDouble(), c11.toDouble(), c12.toDouble(), c13.toDouble(),
			c20.toDouble(), c21.toDouble(), c22.toDouble(), c23.toDouble(),
			c30.toDouble(), c31.toDouble(), c32.toDouble(), c33.toDouble(),
		)
	}

	/**
	 * Creates a copy of this matrix as [MutableMatrix3d],
	 * truncating last row and column
	 */
	fun toMutableMatrix3d(): MutableMatrix3d {
		return MutableMatrix3d(
			c00.toDouble(), c01.toDouble(), c02.toDouble(),
			c10.toDouble(), c11.toDouble(), c12.toDouble(),
			c20.toDouble(), c21.toDouble(), c22.toDouble(),
		)
	}

	/**
	 * Creates a copy of this matrix as [MutableMatrix2d],
	 * truncating two last rows and columns
	 */
	fun toMutableMatrix2d(): MutableMatrix2d {
		return MutableMatrix2d(
			c00.toDouble(), c01.toDouble(),
			c10.toDouble(), c11.toDouble(),
		)
	}

	/**
	 * Creates a copy of this matrix as [Matrix4f]
	 */
	fun toMatrix4f(): Matrix4f {
		return Matrix4f(
			c00, c01, c02, c03,
			c10, c11, c12, c13,
			c20, c21, c22, c23,
			c30, c31, c32, c33,
		)
	}

	/**
	 * Creates a copy of this matrix as [Matrix3f],
	 * truncating last row and column
	 */
	fun toMatrix3f(): Matrix3f {
		return Matrix3f(
			c00, c01, c02,
			c10, c11, c12,
			c20, c21, c22,
		)
	}

	/**
	 * Creates a copy of this matrix as [Matrix2f],
	 * truncating two last rows and columns
	 */
	fun toMatrix2f(): Matrix2f {
		return Matrix2f(
			c00, c01,
			c10, c11,
		)
	}

	/**
	 * Creates a copy of this matrix as [MutableMatrix4f]
	 */
	fun toMutableMatrix4f(): MutableMatrix4f {
		return MutableMatrix4f(
			c00, c01, c02, c03,
			c10, c11, c12, c13,
			c20, c21, c22, c23,
			c30, c31, c32, c33,
		)
	}

	/**
	 * Creates a copy of this matrix as [MutableMatrix3f],
	 * truncating last row and column
	 */
	fun toMutableMatrix3f(): MutableMatrix3f {
		return MutableMatrix3f(
			c00, c01, c02,
			c10, c11, c12,
			c20, c21, c22,
		)
	}

	/**
	 * Creates a copy of this matrix as [MutableMatrix2f],
	 * truncating two last rows and columns
	 */
	fun toMutableMatrix2f(): MutableMatrix2f {
		return MutableMatrix2f(
			c00, c01,
			c10, c11,
		)
	}
}
