
@file:Suppress("unused")

package ru.dbotthepony.kvector.matrix.ndouble

import ru.dbotthepony.kvector.api.*
import ru.dbotthepony.kvector.api.concrete.IMatrix4d
import ru.dbotthepony.kvector.matrix.ndouble.Matrix4d.Companion.ZERO
import ru.dbotthepony.kvector.matrix.ndouble.Matrix4d.Companion.rm
import ru.dbotthepony.kvector.matrix.ndouble.MutableMatrix4d.Companion.rm
import ru.dbotthepony.kvector.narray.Double2Dimensional
import ru.dbotthepony.kvector.vector.ndouble.MutableVector4d
import ru.dbotthepony.kvector.vector.ndouble.Vector3d
import ru.dbotthepony.kvector.vector.ndouble.Vector4d

/**
 * Represents immutable concrete matrix with 4x4 dimensions, storing values as [Double]s.
 *
 * Primary constructor accept values in *Column-Major* order, that is, first row of values
 * is first column in matrix. For *Row-Major* constructor, you should use [rm]
 * companion object method.
 *
 * By default, primary constructor initialize identity matrix, e.g. main diagonal set to 1s.
 *
 * Matrices have multiple sets of properties to access values stored, for convenience of use:
 * - rXY - access row X and column Y
 * - cXY - access column X and row Y
 * - aXY - access row X and column Y, but indices start from 1
 * - bXY - access column X and row Y, but indices start from 1
 * - cX  - access column X as [Vector4d]
 * - rX  - access row X as [Vector4d]
 *
 * **If you don't care about mutability of matrix your method want to accept, use [IMatrix4d]**,
 * because [MutableMatrix4d] is not a subtype of [Matrix4d], but they both are subtype of [IMatrix4d].
 */
class Matrix4d : AbstractMatrixVd<Matrix4d>, IMatrix4d<Matrix4d> {
	constructor(
		c00: Double = 1.0, c01: Double = 0.0, c02: Double = 0.0, c03: Double = 0.0,
		c10: Double = 0.0, c11: Double = 1.0, c12: Double = 0.0, c13: Double = 0.0,
		c20: Double = 0.0, c21: Double = 0.0, c22: Double = 1.0, c23: Double = 0.0,
		c30: Double = 0.0, c31: Double = 0.0, c32: Double = 0.0, c33: Double = 1.0,
	) : super(4, 4, Double2Dimensional(4, 4).also {
		it[0, 0] = c00
		it[1, 0] = c10
		it[2, 0] = c20
		it[3, 0] = c30

		it[0, 1] = c01
		it[1, 1] = c11
		it[2, 1] = c21
		it[3, 1] = c31

		it[0, 2] = c02
		it[1, 2] = c12
		it[2, 2] = c22
		it[3, 2] = c32

		it[0, 3] = c03
		it[1, 3] = c13
		it[2, 3] = c23
		it[3, 3] = c33
	})

	private constructor(memory: Double2Dimensional) : super(4, 4, memory)
	override val flexible: Boolean = false

	override val c00: Double get() = memory[0, 0]
	override val c10: Double get() = memory[1, 0]
	override val c20: Double get() = memory[2, 0]
	override val c30: Double get() = memory[3, 0]
	override val c01: Double get() = memory[0, 1]
	override val c11: Double get() = memory[1, 1]
	override val c21: Double get() = memory[2, 1]
	override val c31: Double get() = memory[3, 1]
	override val c02: Double get() = memory[0, 2]
	override val c12: Double get() = memory[1, 2]
	override val c22: Double get() = memory[2, 2]
	override val c32: Double get() = memory[3, 2]
	override val c03: Double get() = memory[0, 3]
	override val c13: Double get() = memory[1, 3]
	override val c23: Double get() = memory[2, 3]
	override val c33: Double get() = memory[3, 3]

	override val r00 by this::c00
	override val r01 by this::c10
	override val r02 by this::c20
	override val r03 by this::c30
	override val r10 by this::c01
	override val r11 by this::c11
	override val r12 by this::c21
	override val r13 by this::c31
	override val r20 by this::c02
	override val r21 by this::c12
	override val r22 by this::c22
	override val r23 by this::c32
	override val r30 by this::c03
	override val r31 by this::c13
	override val r32 by this::c23
	override val r33 by this::c33

	override val a11 by this::c00
	override val a21 by this::c01
	override val a31 by this::c02
	override val a41 by this::c03
	override val a12 by this::c10
	override val a22 by this::c11
	override val a32 by this::c12
	override val a42 by this::c13
	override val a13 by this::c20
	override val a23 by this::c21
	override val a33 by this::c22
	override val a43 by this::c23
	override val a14 by this::c30
	override val a24 by this::c31
	override val a34 by this::c32
	override val a44 by this::c33

	override val b11 by this::c00
	override val b21 by this::c10
	override val b31 by this::c20
	override val b41 by this::c30
	override val b12 by this::c01
	override val b22 by this::c11
	override val b32 by this::c21
	override val b42 by this::c31
	override val b13 by this::c02
	override val b23 by this::c12
	override val b33 by this::c22
	override val b43 by this::c32
	override val b14 by this::c03
	override val b24 by this::c13
	override val b34 by this::c23
	override val b44 by this::c33

	override val c0: Vector4d get() = Vector4d(c00, c01, c02, c03)
	override val c1: Vector4d get() = Vector4d(c10, c11, c12, c13)
	override val c2: Vector4d get() = Vector4d(c20, c21, c22, c23)
	override val c3: Vector4d get() = Vector4d(c30, c31, c32, c33)

	override val r0: Vector4d get() = Vector4d(r00, r01, r02, r03)
	override val r1: Vector4d get() = Vector4d(r10, r11, r12, r13)
	override val r2: Vector4d get() = Vector4d(r20, r21, r22, r23)
	override val r3: Vector4d get() = Vector4d(r30, r31, r32, r33)

	override fun factorize(memory: Double2Dimensional): Matrix4d {
		requireSizeEquals(memory)
		return Matrix4d(memory)
	}

	override fun translate(vector: Vector3d): Matrix4d {
		return rm(
			r00, r01, r02, r03 + vector.x,
			r10, r11, r12, r13 + vector.y,
			r20, r21, r22, r23 + vector.z,
			r30, r31, r32, r33,
		)
	}

	override fun translateWithMultiplication(vector: Vector3d): Matrix4d {
		return rm(
			r00, r01, r02, r03 + vector.x * r00 + vector.y * r01 + vector.z * r02,
			r10, r11, r12, r13 + vector.x * r10 + vector.y * r11 + vector.z * r12,
			r20, r21, r22, r23 + vector.x * r20 + vector.y * r21 + vector.z * r22,
			r30, r31, r32, r33,
		)
	}

	override fun scale(vector: Vector4d): Matrix4d {
		return rm(
			r00 * vector.x, r01, r02, r03,
			r10, r11 * vector.y, r12, r13,
			r20, r21, r22 * vector.z, r23,
			r30, r31, r32, r33 * vector.w,
		)
	}

	companion object {
		/**
		 * Initializes new matrix, with arguments taken in *Row-Major* order, that is, fist row of values
		 * is fist row in created matrix.
		 */
		fun rm(
			r00: Double = 1.0, r01: Double = 0.0, r02: Double = 0.0, r03: Double = 0.0,
			r10: Double = 0.0, r11: Double = 1.0, r12: Double = 0.0, r13: Double = 0.0,
			r20: Double = 0.0, r21: Double = 0.0, r22: Double = 1.0, r23: Double = 0.0,
			r30: Double = 0.0, r31: Double = 0.0, r32: Double = 0.0, r33: Double = 1.0,
		): Matrix4d {
			return Matrix4d(
				c00 = r00, c10 = r01, c20 = r02, c30 = r03,
				c01 = r10, c11 = r11, c21 = r12, c31 = r13,
				c02 = r20, c12 = r21, c22 = r22, c32 = r23,
				c03 = r30, c13 = r31, c23 = r32, c33 = r33,
			)
		}

		/**
		 * Initializes new zero matrix.
		 *
		 * Since [Matrix4d] is immutable, it is encouraged you use [ZERO] property instead.
		 */
		fun zero(): Matrix4d {
			return Matrix4d(
				c00 = 0.0, c10 = 0.0, c20 = 0.0, c30 = 0.0,
				c01 = 0.0, c11 = 0.0, c21 = 0.0, c31 = 0.0,
				c02 = 0.0, c12 = 0.0, c22 = 0.0, c32 = 0.0,
				c03 = 0.0, c13 = 0.0, c23 = 0.0, c33 = 0.0,
			)
		}

		/**
		 * All values of this matrix are zero
		 */
		@JvmField val ZERO = zero()

		/**
		 * Main diagonal of this matrix is 1s
		 */
		@JvmField val IDENTITY = Matrix4d()
	}
}

/**
 * Represents mutable concrete matrix with 4x4 dimensions, storing values as [Double]s
 *
 * Primary constructor accept values in *Column-Major* order, that is, first row of values
 * is first column in matrix. For *Row-Major* constructor, you should use [rm]
 * companion object method.
 *
 * Same notes about properties of [Matrix4d] apply.
 *
 * **This class is not a subclass of [Matrix4d]**, but a subclass of [IMatrix4d].
 *
 * Differences:
 * - All properties are mutable
 * - Vectorized access return [MutableVector4d], which actually reflect columns/rows of this matrix.
 * That is, any changes done to matrix will be instantly reflected by vector, and any changes to vector
 * will be reflected in matrix.
 * - Vectorized access has setter, and when writing [Vector4d] to it, it will update associated column/row values
 *
 */
class MutableMatrix4d : AbstractMutableMatrixVd<MutableMatrix4d>, IMatrix4d<MutableMatrix4d>, IMutableSquareMatrix<MutableMatrix4d, Vector3d, Vector4d> {
	constructor(
		c00: Double = 1.0, c01: Double = 0.0, c02: Double = 0.0, c03: Double = 0.0,
		c10: Double = 0.0, c11: Double = 1.0, c12: Double = 0.0, c13: Double = 0.0,
		c20: Double = 0.0, c21: Double = 0.0, c22: Double = 1.0, c23: Double = 0.0,
		c30: Double = 0.0, c31: Double = 0.0, c32: Double = 0.0, c33: Double = 1.0,
	) : super(4, 4, Double2Dimensional(4, 4).also {
		it[0, 0] = c00
		it[1, 0] = c10
		it[2, 0] = c20
		it[3, 0] = c30

		it[0, 1] = c01
		it[1, 1] = c11
		it[2, 1] = c21
		it[3, 1] = c31

		it[0, 2] = c02
		it[1, 2] = c12
		it[2, 2] = c22
		it[3, 2] = c32

		it[0, 3] = c03
		it[1, 3] = c13
		it[2, 3] = c23
		it[3, 3] = c33
	})

	private constructor(memory: Double2Dimensional) : super(4, 4, memory)
	override val flexible: Boolean = false

	override fun factorize(memory: Double2Dimensional): MutableMatrix4d {
		requireSizeEquals(memory)
		return MutableMatrix4d(memory)
	}

	override fun translate(vector: Vector3d): MutableMatrix4d {
		return rm(
			r00, r01, r02, r03 + vector.x,
			r10, r11, r12, r13 + vector.y,
			r20, r21, r22, r23 + vector.z,
			r30, r31, r32, r33,
		)
	}

	override fun translateWithMultiplication(vector: Vector3d): MutableMatrix4d {
		return rm(
			r00, r01, r02, r03 + vector.x * r00 + vector.y * r01 + vector.z * r02,
			r10, r11, r12, r13 + vector.x * r10 + vector.y * r11 + vector.z * r12,
			r20, r21, r22, r23 + vector.x * r20 + vector.y * r21 + vector.z * r22,
			r30, r31, r32, r33,
		)
	}

	override fun scale(vector: Vector4d): MutableMatrix4d {
		return rm(
			r00 * vector.x, r01, r02, r03,
			r10, r11 * vector.y, r12, r13,
			r20, r21, r22 * vector.z, r23,
			r30, r31, r32, r33 * vector.w,
		)
	}

	override fun translateWithMultiplicationMut(vector: Vector3d): MutableMatrix4d {
		r03 += vector.x * r00 + vector.y * r01 + vector.z * r02
		r13 += vector.x * r10 + vector.y * r11 + vector.z * r12
		r23 += vector.x * r20 + vector.y * r21 + vector.z * r22
		return this
	}

	override fun scaleMut(vector: Vector4d): MutableMatrix4d {
		c00 *= vector.x
		c11 *= vector.y
		c22 *= vector.z
		c33 *= vector.w
		return this
	}

	override var translation: Vector3d
		get() = super.translation
		set(value) {
			r03 = value.x
			r13 = value.y
			r23 = value.z
		}

	override fun translateMut(vector: Vector3d): MutableMatrix4d {
		r03 += vector.x
		r13 += vector.y
		r23 += vector.z
		return this
	}

	override var c00: Double
		get() = memory[0, 0]
		set(value) { memory[0, 0] = value }
	override var c10: Double
		get() = memory[1, 0]
		set(value) { memory[1, 0] = value }
	override var c20: Double
		get() = memory[2, 0]
		set(value) { memory[2, 0] = value }
	override var c30: Double
		get() = memory[3, 0]
		set(value) { memory[3, 0] = value }
	override var c01: Double
		get() = memory[0, 1]
		set(value) { memory[0, 1] = value }
	override var c11: Double
		get() = memory[1, 1]
		set(value) { memory[1, 1] = value }
	override var c21: Double
		get() = memory[2, 1]
		set(value) { memory[2, 1] = value }
	override var c31: Double
		get() = memory[3, 1]
		set(value) { memory[3, 1] = value }
	override var c02: Double
		get() = memory[0, 2]
		set(value) { memory[0, 2] = value }
	override var c12: Double
		get() = memory[1, 2]
		set(value) { memory[1, 2] = value }
	override var c22: Double
		get() = memory[2, 2]
		set(value) { memory[2, 2] = value }
	override var c32: Double
		get() = memory[3, 2]
		set(value) { memory[3, 2] = value }
	override var c03: Double
		get() = memory[0, 3]
		set(value) { memory[0, 3] = value }
	override var c13: Double
		get() = memory[1, 3]
		set(value) { memory[1, 3] = value }
	override var c23: Double
		get() = memory[2, 3]
		set(value) { memory[2, 3] = value }
	override var c33: Double
		get() = memory[3, 3]
		set(value) { memory[3, 3] = value }

	override var r00 by this::c00
	override var r01 by this::c10
	override var r02 by this::c20
	override var r03 by this::c30
	override var r10 by this::c01
	override var r11 by this::c11
	override var r12 by this::c21
	override var r13 by this::c31
	override var r20 by this::c02
	override var r21 by this::c12
	override var r22 by this::c22
	override var r23 by this::c32
	override var r30 by this::c03
	override var r31 by this::c13
	override var r32 by this::c23
	override var r33 by this::c33

	override var a11 by this::c00
	override var a21 by this::c01
	override var a31 by this::c02
	override var a41 by this::c03
	override var a12 by this::c10
	override var a22 by this::c11
	override var a32 by this::c12
	override var a42 by this::c13
	override var a13 by this::c20
	override var a23 by this::c21
	override var a33 by this::c22
	override var a43 by this::c23
	override var a14 by this::c30
	override var a24 by this::c31
	override var a34 by this::c32
	override var a44 by this::c33

	override var b11 by this::c00
	override var b21 by this::c01
	override var b31 by this::c02
	override var b41 by this::c03
	override var b12 by this::c10
	override var b22 by this::c11
	override var b32 by this::c12
	override var b42 by this::c13
	override var b13 by this::c20
	override var b23 by this::c21
	override var b33 by this::c22
	override var b43 by this::c23
	override var b14 by this::c30
	override var b24 by this::c31
	override var b34 by this::c32
	override var b44 by this::c33

	override var c0: Vector4d
		get() = object : MutableVector4d(c00, c01, c02, c03) {
			override var x by this@MutableMatrix4d::c00
			override var y by this@MutableMatrix4d::c01
			override var z by this@MutableMatrix4d::c02
			override var w by this@MutableMatrix4d::c03
		}
		set(value) {
			c00 = value.x
			c01 = value.y
			c02 = value.z
			c03 = value.w
		}

	override var c1: Vector4d
		get() = object : MutableVector4d(c10, c11, c12, c13) {
			override var x by this@MutableMatrix4d::c10
			override var y by this@MutableMatrix4d::c11
			override var z by this@MutableMatrix4d::c12
			override var w by this@MutableMatrix4d::c13
		}
		set(value) {
			c10 = value.x
			c11 = value.y
			c12 = value.z
			c13 = value.w
		}

	override var c2: Vector4d
		get() = object : MutableVector4d(c20, c21, c22, c23) {
			override var x by this@MutableMatrix4d::c20
			override var y by this@MutableMatrix4d::c21
			override var z by this@MutableMatrix4d::c22
			override var w by this@MutableMatrix4d::c23
		}
		set(value) {
			c20 = value.x
			c21 = value.y
			c22 = value.z
			c23 = value.w
		}

	override var c3: Vector4d
		get() = object : MutableVector4d(c30, c31, c32, c33) {
			override var x by this@MutableMatrix4d::c30
			override var y by this@MutableMatrix4d::c31
			override var z by this@MutableMatrix4d::c32
			override var w by this@MutableMatrix4d::c33
		}
		set(value) {
			c30 = value.x
			c31 = value.y
			c32 = value.z
			c33 = value.w
		}

	override var r0: Vector4d
		get() = object : MutableVector4d(r00, r01, r02, r03) {
			override var x by this@MutableMatrix4d::r00
			override var y by this@MutableMatrix4d::r01
			override var z by this@MutableMatrix4d::r02
			override var w by this@MutableMatrix4d::r03
		}
		set(value) {
			r00 = value.x
			r01 = value.y
			r02 = value.z
			r03 = value.w
		}

	override var r1: Vector4d
		get() = object : MutableVector4d(r10, r11, r12, r13) {
			override var x by this@MutableMatrix4d::r10
			override var y by this@MutableMatrix4d::r11
			override var z by this@MutableMatrix4d::r12
			override var w by this@MutableMatrix4d::r13
		}
		set(value) {
			r10 = value.x
			r11 = value.y
			r12 = value.z
			r13 = value.w
		}

	override var r2: Vector4d
		get() = object : MutableVector4d(r20, r21, r22, r23) {
			override var x by this@MutableMatrix4d::r20
			override var y by this@MutableMatrix4d::r21
			override var z by this@MutableMatrix4d::r22
			override var w by this@MutableMatrix4d::r23
		}
		set(value) {
			r20 = value.x
			r21 = value.y
			r22 = value.z
			r23 = value.w
		}

	override var r3: Vector4d
		get() = object : MutableVector4d(r30, r31, r32, r33) {
			override var x by this@MutableMatrix4d::r30
			override var y by this@MutableMatrix4d::r31
			override var z by this@MutableMatrix4d::r32
			override var w by this@MutableMatrix4d::r33
		}
		set(value) {
			r30 = value.x
			r31 = value.y
			r32 = value.z
			r33 = value.w
		}

	companion object {
		/**
		 * Initializes new matrix, with arguments taken in *Row-Major* order, that is, fist row of values
		 * is fist row in created matrix.
		 */
		fun rm(
			r00: Double = 1.0, r01: Double = 0.0, r02: Double = 0.0, r03: Double = 0.0,
			r10: Double = 0.0, r11: Double = 1.0, r12: Double = 0.0, r13: Double = 0.0,
			r20: Double = 0.0, r21: Double = 0.0, r22: Double = 1.0, r23: Double = 0.0,
			r30: Double = 0.0, r31: Double = 0.0, r32: Double = 0.0, r33: Double = 1.0,
		): MutableMatrix4d {
			return MutableMatrix4d(
				c00 = r00, c10 = r01, c20 = r02, c30 = r03,
				c01 = r10, c11 = r11, c21 = r12, c31 = r13,
				c02 = r20, c12 = r21, c22 = r22, c32 = r23,
				c03 = r30, c13 = r31, c23 = r32, c33 = r33,
			)
		}

		/**
		 * Initializes new mutable zero matrix.
		 */
		fun zero(): MutableMatrix4d {
			return MutableMatrix4d(
				c00 = 0.0, c10 = 0.0, c20 = 0.0, c30 = 0.0,
				c01 = 0.0, c11 = 0.0, c21 = 0.0, c31 = 0.0,
				c02 = 0.0, c12 = 0.0, c22 = 0.0, c32 = 0.0,
				c03 = 0.0, c13 = 0.0, c23 = 0.0, c33 = 0.0,
			)
		}
	}
}
