
@file:Suppress("unused")

package ru.dbotthepony.kvector.matrix.nfloat

import ru.dbotthepony.kvector.api.IMutableSquareMatrix
import ru.dbotthepony.kvector.api.concrete.IMatrix3f
import ru.dbotthepony.kvector.matrix.ndouble.Matrix4d
import ru.dbotthepony.kvector.matrix.ndouble.MutableMatrix4d
import ru.dbotthepony.kvector.narray.Float2Dimensional
import ru.dbotthepony.kvector.vector.nfloat.*

/**
 * Represents immutable concrete matrix with 3x3 dimensions, storing values as [Float]s.
 *
 * See [Matrix4d] for documentation, with reflection of this class.
 *
 * Vectorized access use [Vector3f], generic interface is represented by [IMatrix3f].
 */
class Matrix3f : AbstractMatrixVf<Matrix3f>, IMatrix3f<Matrix3f> {
	constructor(
		c00: Float = 1f, c01: Float = 0f, c02: Float = 0f,
		c10: Float = 0f, c11: Float = 1f, c12: Float = 0f,
		c20: Float = 0f, c21: Float = 0f, c22: Float = 1f,
	) : super(3, 3, Float2Dimensional(3, 3).also {
		it[0, 0] = c00
		it[1, 0] = c10
		it[2, 0] = c20

		it[0, 1] = c01
		it[1, 1] = c11
		it[2, 1] = c21

		it[0, 2] = c02
		it[1, 2] = c12
		it[2, 2] = c22
	})

	private constructor(memory: Float2Dimensional) : super(3, 3, memory)

	override val flexible: Boolean = false

	override fun factorize(memory: Float2Dimensional): Matrix3f {
		requireSizeEquals(memory)
		return Matrix3f(memory)
	}

	override val c00: Float get() = memory[0, 0]
	override val c10: Float get() = memory[1, 0]
	override val c20: Float get() = memory[2, 0]
	override val c01: Float get() = memory[0, 1]
	override val c11: Float get() = memory[1, 1]
	override val c21: Float get() = memory[2, 1]
	override val c02: Float get() = memory[0, 2]
	override val c12: Float get() = memory[1, 2]
	override val c22: Float get() = memory[2, 2]

	override val r00 by this::c00
	override val r01 by this::c10
	override val r02 by this::c20
	override val r10 by this::c01
	override val r11 by this::c11
	override val r12 by this::c21
	override val r20 by this::c02
	override val r21 by this::c12
	override val r22 by this::c22

	override val a11 by this::c00
	override val a21 by this::c01
	override val a31 by this::c02
	override val a12 by this::c10
	override val a22 by this::c11
	override val a32 by this::c12
	override val a13 by this::c20
	override val a23 by this::c21
	override val a33 by this::c22

	override val b11 by this::c00
	override val b21 by this::c10
	override val b31 by this::c20
	override val b12 by this::c01
	override val b22 by this::c11
	override val b32 by this::c21
	override val b13 by this::c02
	override val b23 by this::c12
	override val b33 by this::c22

	override val c0: Vector3f get() = Vector3f(c00, c01, c02)
	override val c1: Vector3f get() = Vector3f(c10, c11, c12)
	override val c2: Vector3f get() = Vector3f(c20, c21, c22)

	override val r0: Vector3f get() = Vector3f(r00, r01, r02)
	override val r1: Vector3f get() = Vector3f(r10, r11, r12)
	override val r2: Vector3f get() = Vector3f(r20, r21, r22)

	override fun translate(vector: Vector2f): Matrix3f {
		return rm(
			r00, r01, r02 + vector.x,
			r10, r11, r12 + vector.y,
			r20, r21, r22,
		)
	}

	override fun translateWithMultiplication(vector: Vector2f): Matrix3f {
		return rm(
			r00, r01, r02 + vector.x * r00 + vector.y * r01,
			r10, r11, r12 + vector.x * r10 + vector.y * r11,
			r20, r21, r22,
		)
	}

	override fun scale(vector: Vector3f): Matrix3f {
		return rm(
			r00 * vector.x, r01, r02,
			r10, r11 * vector.y, r12,
			r20, r21, r22 * vector.z
		)
	}

	companion object {
		/**
		 * Initializes new matrix, with arguments taken in *Row-Major* order, that is, fist row of values
		 * is fist row in created matrix.
		 */
		fun rm(
			r00: Float = 1f, r01: Float = 0f, r02: Float = 0f,
			r10: Float = 0f, r11: Float = 1f, r12: Float = 0f,
			r20: Float = 0f, r21: Float = 0f, r22: Float = 1f,
		): Matrix3f {
			return Matrix3f(
				c00 = r00, c10 = r01, c20 = r02,
				c01 = r10, c11 = r11, c21 = r12,
				c02 = r20, c12 = r21, c22 = r22,
			)
		}

		/**
		 * Initializes new zero matrix.
		 *
		 * Since [Matrix3f] is immutable, it is encouraged you use [ZERO] property instead.
		 */
		fun zero(): Matrix3f {
			return Matrix3f(
				c00 = 0f, c10 = 0f, c20 = 0f,
				c01 = 0f, c11 = 0f, c21 = 0f,
				c02 = 0f, c12 = 0f, c22 = 0f,
			)
		}

		/**
		 * All values of this matrix are zero
		 */
		@JvmField val ZERO = zero()

		/**
		 * Main diagonal of this matrix is 1s
		 */
		@JvmField val IDENTITY = Matrix3f()
	}
}

/**
 * Represents mutable concrete matrix with 3x3 dimensions, storing values as [Float]s.
 *
 * See [MutableMatrix4d] for documentation, with reflection of this class.
 *
 * Vectorized access use [MutableVector3f], generic interface is represented by [IMatrix3f].
 */
class MutableMatrix3f : AbstractMutableMatrixVf<MutableMatrix3f>, IMutableSquareMatrix<MutableMatrix3f, Vector2f, Vector3f>, IMatrix3f<MutableMatrix3f> {
	constructor(
		c00: Float = 1f, c01: Float = 0f, c02: Float = 0f,
		c10: Float = 0f, c11: Float = 1f, c12: Float = 0f,
		c20: Float = 0f, c21: Float = 0f, c22: Float = 1f,
	) : super(3, 3, Float2Dimensional(3, 3).also {
		it[0, 0] = c00
		it[1, 0] = c10
		it[2, 0] = c20

		it[0, 1] = c01
		it[1, 1] = c11
		it[2, 1] = c21

		it[0, 2] = c02
		it[1, 2] = c12
		it[2, 2] = c22
	})

	private constructor(memory: Float2Dimensional) : super(3, 3, memory)

	override val flexible: Boolean = false

	override fun factorize(memory: Float2Dimensional): MutableMatrix3f {
		requireSizeEquals(memory)
		return MutableMatrix3f(memory)
	}

	override fun translateWithMultiplicationMut(vector: Vector2f): MutableMatrix3f {
		r02 += vector.x * r00 + vector.y * r01
		r12 += vector.x * r10 + vector.y * r11
		return this
	}

	override fun scaleMut(vector: Vector3f): MutableMatrix3f {
		r00 *= vector.x
		r11 *= vector.y
		r22 *= vector.z
		return this
	}

	override fun translateMut(vector: Vector2f): MutableMatrix3f {
		r02 += vector.x
		r12 += vector.y
		return this
	}

	override var c00: Float
		get() = memory[0, 0]
		set(value) { memory[0, 0] = value }
	override var c10: Float
		get() = memory[1, 0]
		set(value) { memory[1, 0] = value }
	override var c20: Float
		get() = memory[2, 0]
		set(value) { memory[2, 0] = value }
	override var c01: Float
		get() = memory[0, 1]
		set(value) { memory[0, 1] = value }
	override var c11: Float
		get() = memory[1, 1]
		set(value) { memory[1, 1] = value }
	override var c21: Float
		get() = memory[2, 1]
		set(value) { memory[2, 1] = value }
	override var c02: Float
		get() = memory[0, 2]
		set(value) { memory[0, 2] = value }
	override var c12: Float
		get() = memory[1, 2]
		set(value) { memory[1, 2] = value }
	override var c22: Float
		get() = memory[2, 2]
		set(value) { memory[2, 2] = value }

	override var r00 by this::c00
	override var r01 by this::c10
	override var r02 by this::c20
	override var r10 by this::c01
	override var r11 by this::c11
	override var r12 by this::c21
	override var r20 by this::c02
	override var r21 by this::c12
	override var r22 by this::c22

	override var a11 by this::c00
	override var a21 by this::c01
	override var a31 by this::c02
	override var a12 by this::c10
	override var a22 by this::c11
	override var a32 by this::c12
	override var a13 by this::c20
	override var a23 by this::c21
	override var a33 by this::c22

	override var b11 by this::c00
	override var b21 by this::c10
	override var b31 by this::c20
	override var b12 by this::c01
	override var b22 by this::c11
	override var b32 by this::c21
	override var b13 by this::c02
	override var b23 by this::c12
	override var b33 by this::c22

	override var translation: Vector2f
		get() = super.translation
		set(value) {
			r02 = value.x
			r12 = value.y
		}

	override var c0: Vector3f
		get() = object : MutableVector3f(c00, c01, c02) {
			override var x by this@MutableMatrix3f::c00
			override var y by this@MutableMatrix3f::c01
			override var z by this@MutableMatrix3f::c02
		}
		set(value) {
			c00 = value.x
			c01 = value.y
			c02 = value.z
		}
	override var c1: Vector3f
		get() = object : MutableVector3f(c10, c11, c12) {
			override var x by this@MutableMatrix3f::c10
			override var y by this@MutableMatrix3f::c11
			override var z by this@MutableMatrix3f::c12
		}
		set(value) {
			c10 = value.x
			c11 = value.y
			c12 = value.z
		}
	override var c2: Vector3f
		get() = object : MutableVector3f(c20, c21, c22) {
			override var x by this@MutableMatrix3f::c20
			override var y by this@MutableMatrix3f::c21
			override var z by this@MutableMatrix3f::c22
		}
		set(value) {
			c20 = value.x
			c21 = value.y
			c22 = value.z
		}

	override var r0: Vector3f
		get() = object : MutableVector3f(r00, r01, r02) {
			override var x by this@MutableMatrix3f::r00
			override var y by this@MutableMatrix3f::r01
			override var z by this@MutableMatrix3f::r02
		}
		set(value) {
			r00 = value.x
			r01 = value.y
			r02 = value.z
		}
	override var r1: Vector3f
		get() = object : MutableVector3f(r10, r11, r12) {
			override var x by this@MutableMatrix3f::r10
			override var y by this@MutableMatrix3f::r11
			override var z by this@MutableMatrix3f::r12
		}
		set(value) {
			r10 = value.x
			r11 = value.y
			r12 = value.z
		}
	override var r2: Vector3f
		get() = object : MutableVector3f(r20, r21, r22) {
			override var x by this@MutableMatrix3f::r20
			override var y by this@MutableMatrix3f::r21
			override var z by this@MutableMatrix3f::r22
		}
		set(value) {
			r20 = value.x
			r21 = value.y
			r22 = value.z
		}

	override fun translate(vector: Vector2f): MutableMatrix3f {
		return rm(
			r00, r01, r02 + vector.x,
			r10, r11, r12 + vector.y,
			r20, r21, r22,
		)
	}

	override fun translateWithMultiplication(vector: Vector2f): MutableMatrix3f {
		return rm(
			r00, r01, r02 + vector.x * r00 + vector.y * r01,
			r10, r11, r12 + vector.x * r10 + vector.y * r11,
			r20, r21, r22,
		)
	}

	override fun scale(vector: Vector3f): MutableMatrix3f {
		return rm(
			r00 * vector.x, r01, r02,
			r10, r11 * vector.y, r12,
			r20, r21, r22 * vector.z
		)
	}

	companion object {
		/**
		 * Initializes new matrix, with arguments taken in *Row-Major* order, that is, fist row of values
		 * is fist row in created matrix.
		 */
		fun rm(
			r00: Float = 1f, r01: Float = 0f, r02: Float = 0f,
			r10: Float = 0f, r11: Float = 1f, r12: Float = 0f,
			r20: Float = 0f, r21: Float = 0f, r22: Float = 1f,
		): MutableMatrix3f {
			return MutableMatrix3f(
				c00 = r00, c10 = r01, c20 = r02,
				c01 = r10, c11 = r11, c21 = r12,
				c02 = r20, c12 = r21, c22 = r22,
			)
		}

		/**
		 * Initializes new zero matrix.
		 */
		fun zero(): MutableMatrix3f {
			return MutableMatrix3f(
				c00 = 0f, c10 = 0f, c20 = 0f,
				c01 = 0f, c11 = 0f, c21 = 0f,
				c02 = 0f, c12 = 0f, c22 = 0f,
			)
		}
	}
}
