
@file:Suppress("nothing_to_inline", "unused")

package ru.dbotthepony.kvector.narray

import ru.dbotthepony.kvector.api.*

private inline fun require(condition: Boolean, lazy: () -> Any) {
	if (!condition) {
		throw IndexOutOfBoundsException(lazy.invoke().toString())
	}
}

/**
 * Two-dimensional array of [Double]s, backed by primitive [DoubleArray].
 *
 * Has two constructors, one is for constructing fresh two-dimensional array, and second is for
 * viewing already existing array as two-dimensional array.
 */
class Double2Dimensional(
	override val columns: Int,
	override val rows: Int,
	@JvmField val memory: DoubleArray,
	private val offset: Int = 0,
) : IMatrixGetterDouble, IMatrixSetterDouble {
	init {
		require(columns >= 0) { "Tried to create $columns columns" }
		require(rows >= 0) { "Tried to create $rows rows" }
		require(memory.size - offset >= columns * rows) { "Provided array (${memory.size}) does not satisfy demands of $offset offset, $columns columns and $rows rows" }
	}

	constructor(columns: Int, rows: Int) : this(columns, rows, DoubleArray(columns * rows))

	override operator fun get(column: Int, row: Int): Double {
		require(column in 0 until columns) { "Column out of bounds: $column" }
		require(row in 0 until rows) { "Row out of bounds: $row" }
		return memory[offset + column + row * columns]
	}

	override operator fun set(column: Int, row: Int, value: Double) {
		require(column in 0 until columns) { "Column out of bounds: $column" }
		require(row in 0 until rows) { "Row out of bounds: $row" }
		memory[offset + column + row * columns] = value
	}
}

/**
 * Two-dimensional array of [Float]s, backed by primitive [FloatArray]
 *
 * Has two constructors, one is for constructing fresh two-dimensional array, and second is for
 * viewing already existing array as two-dimensional array.
 */
class Float2Dimensional(
	override val columns: Int,
	override val rows: Int,
	@JvmField val memory: FloatArray,
	private val offset: Int = 0,
) : IMatrixGetterFloat, IMatrixSetterFloat {
	init {
		require(columns >= 0) { "Tried to create $columns columns" }
		require(rows >= 0) { "Tried to create $rows rows" }
		require(memory.size - offset >= columns * rows) { "Provided array (${memory.size}) does not satisfy demands of $offset offset, $columns columns and $rows rows" }
	}

	constructor(columns: Int, rows: Int) : this(columns, rows, FloatArray(columns * rows))

	override operator fun get(column: Int, row: Int): Float {
		require(column in 0 until columns) { "Column out of bounds: $column" }
		require(row in 0 until rows) { "Row out of bounds: $row" }
		return memory[offset + column + row * columns]
	}

	override operator fun set(column: Int, row: Int, value: Float) {
		require(column in 0 until columns) { "Column out of bounds: $column" }
		require(row in 0 until rows) { "Row out of bounds: $row" }
		memory[offset + column + row * columns] = value
	}
}

/**
 * Two-dimensional array of [Byte]s, backed by primitive [ByteArray]
 *
 * Has two constructors, one is for constructing fresh two-dimensional array, and second is for
 * viewing already existing array as two-dimensional array.
 */
class Byte2Dimensional(
	override val columns: Int,
	override val rows: Int,
	@JvmField val memory: ByteArray,
	private val offset: Int = 0,
) : IMatrixGetterByte, IMatrixSetterByte {
	init {
		require(columns >= 0) { "Tried to create $columns columns" }
		require(rows >= 0) { "Tried to create $rows rows" }
		require(memory.size - offset >= columns * rows) { "Provided array (${memory.size}) does not satisfy demands of $offset offset, $columns columns and $rows rows" }
	}

	constructor(columns: Int, rows: Int) : this(columns, rows, ByteArray(columns * rows))

	override operator fun get(column: Int, row: Int): Byte {
		require(column in 0 until columns) { "Column out of bounds: $column" }
		require(row in 0 until rows) { "Row out of bounds: $row" }
		return memory[offset + column + row * columns]
	}

	override operator fun set(column: Int, row: Int, value: Byte) {
		require(column in 0 until columns) { "Column out of bounds: $column" }
		require(row in 0 until rows) { "Row out of bounds: $row" }
		memory[offset + column + row * columns] = value
	}
}

/**
 * Two-dimensional array of [Short]s, backed by primitive [ShortArray]
 *
 * Has two constructors, one is for constructing fresh two-dimensional array, and second is for
 * viewing already existing array as two-dimensional array.
 */
class Short2Dimensional(
	override val columns: Int,
	override val rows: Int,
	@JvmField val memory: ShortArray,
	private val offset: Int = 0,
) : IMatrixGetterShort, IMatrixSetterShort {
	init {
		require(columns >= 0) { "Tried to create $columns columns" }
		require(rows >= 0) { "Tried to create $rows rows" }
		require(memory.size - offset >= columns * rows) { "Provided array (${memory.size}) does not satisfy demands of $offset offset, $columns columns and $rows rows" }
	}

	constructor(columns: Int, rows: Int) : this(columns, rows, ShortArray(columns * rows))

	override operator fun get(column: Int, row: Int): Short {
		require(column in 0 until columns) { "Column out of bounds: $column" }
		require(row in 0 until rows) { "Row out of bounds: $row" }
		return memory[offset + column + row * columns]
	}

	override operator fun set(column: Int, row: Int, value: Short) {
		require(column in 0 until columns) { "Column out of bounds: $column" }
		require(row in 0 until rows) { "Row out of bounds: $row" }
		memory[offset + column + row * columns] = value
	}
}

/**
 * Two-dimensional array of [Int]s, backed by primitive [IntArray]
 *
 * Has two constructors, one is for constructing fresh two-dimensional array, and second is for
 * viewing already existing array as two-dimensional array.
 */
class Int2Dimensional(
	override val columns: Int,
	override val rows: Int,
	@JvmField val memory: IntArray,
	private val offset: Int = 0,
) : IMatrixGetterInt, IMatrixSetterInt {
	init {
		require(columns >= 0) { "Tried to create $columns columns" }
		require(rows >= 0) { "Tried to create $rows rows" }
		require(memory.size - offset >= columns * rows) { "Provided array (${memory.size}) does not satisfy demands of $offset offset, $columns columns and $rows rows" }
	}

	constructor(columns: Int, rows: Int) : this(columns, rows, IntArray(columns * rows))

	override operator fun get(column: Int, row: Int): Int {
		require(column in 0 until columns) { "Column out of bounds: $column" }
		require(row in 0 until rows) { "Row out of bounds: $row" }
		return memory[offset + column + row * columns]
	}

	override operator fun set(column: Int, row: Int, value: Int) {
		require(column in 0 until columns) { "Column out of bounds: $column" }
		require(row in 0 until rows) { "Row out of bounds: $row" }
		memory[offset + column + row * columns] = value
	}
}

/**
 * Two-dimensional array of [Long]s, backed by primitive [LongArray]
 *
 * Has two constructors, one is for constructing fresh two-dimensional array, and second is for
 * viewing already existing array as two-dimensional array.
 */
class Long2Dimensional(
	override val columns: Int,
	override val rows: Int,
	@JvmField val memory: LongArray,
	private val offset: Int = 0,
) : IMatrixGetterLong, IMatrixSetterLong {
	init {
		require(columns >= 0) { "Tried to create $columns columns" }
		require(rows >= 0) { "Tried to create $rows rows" }
		require(memory.size - offset >= columns * rows) { "Provided array (${memory.size}) does not satisfy demands of $offset offset, $columns columns and $rows rows" }
	}

	constructor(columns: Int, rows: Int) : this(columns, rows, LongArray(columns * rows))

	override operator fun get(column: Int, row: Int): Long {
		require(column in 0 until columns) { "Column out of bounds: $column" }
		require(row in 0 until rows) { "Row out of bounds: $row" }
		return memory[offset + column + row * columns]
	}

	override operator fun set(column: Int, row: Int, value: Long) {
		require(column in 0 until columns) { "Column out of bounds: $column" }
		require(row in 0 until rows) { "Row out of bounds: $row" }
		memory[offset + column + row * columns] = value
	}
}
