package ru.dbotthepony.kstarbound.world

import ru.dbotthepony.kstarbound.METRES_IN_STARBOUND_UNIT
import ru.dbotthepony.kstarbound.defs.TileDefinition
import ru.dbotthepony.kstarbound.math.*
import ru.dbotthepony.kstarbound.world.entities.CollisionResolution
import ru.dbotthepony.kstarbound.world.entities.Entity
import ru.dbotthepony.kvector.util2d.AABB
import ru.dbotthepony.kvector.util2d.AABBi
import ru.dbotthepony.kvector.vector.ndouble.Vector2d
import ru.dbotthepony.kvector.vector.nint.Vector2i
import kotlin.math.absoluteValue

/**
 * Кортеж чанка, который содержит родителя (мир) и соседей (кортежи чанков)
 */
interface IWorldChunkTuple<WorldType : World<WorldType, ChunkType>, ChunkType : Chunk<WorldType, ChunkType>> {
	val world: WorldType
	val chunk: ChunkType
	val top: IWorldChunkTuple<WorldType, ChunkType>?
	val left: IWorldChunkTuple<WorldType, ChunkType>?
	val right: IWorldChunkTuple<WorldType, ChunkType>?
	val bottom: IWorldChunkTuple<WorldType, ChunkType>?
}

interface IMutableWorldChunkTuple<WorldType : World<WorldType, ChunkType>, ChunkType : Chunk<WorldType, ChunkType>> : IWorldChunkTuple<WorldType, ChunkType> {
	override var top: IMutableWorldChunkTuple<WorldType, ChunkType>?
	override var left: IMutableWorldChunkTuple<WorldType, ChunkType>?
	override var right: IMutableWorldChunkTuple<WorldType, ChunkType>?
	override var bottom: IMutableWorldChunkTuple<WorldType, ChunkType>?
}

class WorldChunkTuple<WorldType : World<WorldType, ChunkType>, ChunkType : Chunk<WorldType, ChunkType>>(
	private val parent: IWorldChunkTuple<WorldType, ChunkType>
) : IWorldChunkTuple<WorldType, ChunkType> {
	override val world get() = parent.world
	override val chunk get() = parent.chunk
	override val top: IWorldChunkTuple<WorldType, ChunkType>? get() {
		val getValue = parent.top

		if (getValue != null) {
			return WorldChunkTuple(getValue)
		}

		return null
	}

	override val left: IWorldChunkTuple<WorldType, ChunkType>? get() {
		val getValue = parent.left

		if (getValue != null) {
			return WorldChunkTuple(getValue)
		}

		return null
	}

	override val right: IWorldChunkTuple<WorldType, ChunkType>? get() {
		val getValue = parent.right

		if (getValue != null) {
			return WorldChunkTuple(getValue)
		}

		return null
	}

	override val bottom: IWorldChunkTuple<WorldType, ChunkType>? get() {
		val getValue = parent.bottom

		if (getValue != null) {
			return WorldChunkTuple(getValue)
		}

		return null
	}
}

open class MutableWorldChunkTuple<WorldType : World<WorldType, ChunkType>, ChunkType : Chunk<WorldType, ChunkType>>(
	override val world: WorldType,
	override val chunk: ChunkType,
	override var top: IMutableWorldChunkTuple<WorldType, ChunkType>?,
	override var left: IMutableWorldChunkTuple<WorldType, ChunkType>?,
	override var right: IMutableWorldChunkTuple<WorldType, ChunkType>?,
	override var bottom: IMutableWorldChunkTuple<WorldType, ChunkType>?,
) : IMutableWorldChunkTuple<WorldType, ChunkType>

const val EARTH_FREEFALL_ACCELERATION = 9.8312 / METRES_IN_STARBOUND_UNIT

data class WorldSweepResult(
	val hitPosition: Vector2d,
	val collisionTime: Double,
	val hitAnything: Boolean,
	val hitNormal: Vector2d
)

private const val EPSILON = 0.00001

abstract class World<This : World<This, ChunkType>, ChunkType : Chunk<This, ChunkType>>(val seed: Long = 0L) {
	protected val chunkMap = HashMap<ChunkPos, IMutableWorldChunkTuple<This, ChunkType>>()
	protected var lastAccessedChunk: IMutableWorldChunkTuple<This, ChunkType>? = null

	/**
	 * Таймер этого мира, в секундах.
	 *
	 * Условия:
	 * * Это значение МОЖЕТ уменьшаться, в том числе в силу того, что клиент
	 * обсчитывает мир быстрее, чем сервер, и наоборот.
	 * * Точка отсчёта не имеет смысла (по умолчанию отсчитывается от нуля).
	 * * Все сущности обязаны использовать этот таймер для определения времени своих событий
	 * * Таймер НЕ МОЖЕТ возвращать разные значения в пределах одного тика.
	 * т.е. таймер должен быть обновлён только после или перед тиком всех потомков
	 * * НЕ ОБЯЗАН соответствовать течению реального времени, но как правило, соответствует ему
	 * (если позволяет скорость обсчёта потомков).
	 * * ОБЯЗАН быть сохранён при сохранении мира, и выставляться при его загрузке.
	 */
	var timer: Double = 0.0
		protected set

	/**
	 * Заставляет этот мир "думать" на заданное количество секунд.
	 * Вызывает все необходимые методы у потомков, обновляя игровое состояние.
	 */
	fun think(delta: Double) {
		require(delta > 0.0) { "Tried to update $this by $delta seconds" }

		timer += delta
		thinkInner(delta)
	}

	protected abstract fun thinkInner(delta: Double)

	/**
	 * Сущности, у которых нет чанка, но они находятся в этом мире
	 */
	val orphanedEntities = HashSet<Entity>()

	/**
	 * Сущности, находящиеся в этом мире
	 */
	val entities = HashSet<Entity>()

	/**
	 * Стандартное ускорение свободного падения в Starbound Units/секунда^2
	 *
	 * При Vector2d.ZERO = невесомость
	 */
	var gravity = Vector2d(0.0, -EARTH_FREEFALL_ACCELERATION)

	protected abstract fun chunkFactory(
		pos: ChunkPos,
	): ChunkType

	protected fun getChunkInternal(pos: ChunkPos): IMutableWorldChunkTuple<This, ChunkType>? {
		if (lastAccessedChunk?.chunk?.pos == pos) {
			return lastAccessedChunk
		}

		return chunkMap[pos]
	}

	open fun getChunk(pos: ChunkPos): IWorldChunkTuple<This, ChunkType>? {
		val getTuple = getChunkInternal(pos)

		if (getTuple != null)
			return WorldChunkTuple(getTuple)

		return null
	}

	protected open fun computeIfAbsentInternal(pos: ChunkPos): IWorldChunkTuple<This, ChunkType> {
		if (lastAccessedChunk?.chunk?.pos == pos) {
			return lastAccessedChunk!!
		}

		return chunkMap.computeIfAbsent(pos) lazy@{
			val chunk = chunkFactory(pos)

			val top = getChunkInternal(pos.up())
			val left = getChunkInternal(pos.left())
			val right = getChunkInternal(pos.right())
			val bottom = getChunkInternal(pos.down())

			val tuple = MutableWorldChunkTuple(
				world = this as This,
				chunk = chunk,
				top = top,
				left = left,
				right = right,
				bottom = bottom,
			)

			top?.bottom = tuple
			left?.right = tuple
			right?.left = tuple
			bottom?.top = tuple

			lastAccessedChunk = tuple

			val orphanedInThisChunk = ArrayList<Entity>()

			for (ent in orphanedEntities) {
				val cPos = ChunkPos.fromTilePosition(ent.pos)

				if (cPos == pos) {
					orphanedInThisChunk.add(ent)
				}
			}

			for (ent in orphanedInThisChunk) {
				ent.chunk = chunk
			}

			return@lazy tuple
		}
	}

	open fun computeIfAbsent(pos: ChunkPos): IWorldChunkTuple<This, ChunkType> {
		return WorldChunkTuple(computeIfAbsentInternal(pos))
	}

	open fun getForegroundView(pos: ChunkPos): TileChunkView? {
		val get = getChunkInternal(pos) ?: return null

		return TileChunkView(
			center = 			get.chunk.foreground,
			left = 				get.left?.chunk?.foreground,
			top = 				get.top?.chunk?.foreground,
			topLeft = 			getChunkInternal(pos.up().left())?.chunk?.foreground,
			topRight = 			getChunkInternal(pos.up().right())?.chunk?.foreground,
			right = 			get.right?.chunk?.foreground,
			bottom = 			get.bottom?.chunk?.foreground,
			bottomLeft = 		getChunkInternal(pos.down().left())?.chunk?.foreground,
			bottomRight = 		getChunkInternal(pos.down().right())?.chunk?.foreground,
		)
	}

	open fun getBackgroundView(pos: ChunkPos): TileChunkView? {
		val get = getChunkInternal(pos) ?: return null

		return TileChunkView(
			center = 			get.chunk.background,
			left = 				get.left?.chunk?.background,
			top = 				get.top?.chunk?.background,
			topLeft = 			getChunkInternal(pos.up().left())?.chunk?.background,
			topRight = 			getChunkInternal(pos.up().right())?.chunk?.background,
			right = 			get.right?.chunk?.background,
			bottom = 			get.bottom?.chunk?.background,
			bottomLeft = 		getChunkInternal(pos.down().left())?.chunk?.background,
			bottomRight = 		getChunkInternal(pos.down().right())?.chunk?.background,
		)
	}

	fun getTile(pos: Vector2i): ChunkTile? {
		return getChunkInternal(ChunkPos.fromTilePosition(pos))?.chunk?.foreground?.get(ChunkPos.normalizeCoordinate(pos.x), ChunkPos.normalizeCoordinate(pos.y))
	}

	fun setTile(pos: Vector2i, tile: TileDefinition?): IWorldChunkTuple<This, ChunkType> {
		val chunk = computeIfAbsentInternal(ChunkPos.fromTilePosition(pos))
		chunk.chunk.foreground[ChunkPos.normalizeCoordinate(pos.x), ChunkPos.normalizeCoordinate(pos.y)] = tile
		return chunk
	}

	fun getBackgroundTile(pos: Vector2i): ChunkTile? {
		return getChunkInternal(ChunkPos.fromTilePosition(pos))?.chunk?.background?.get(ChunkPos.normalizeCoordinate(pos.x), ChunkPos.normalizeCoordinate(pos.y))
	}

	fun setBackgroundTile(pos: Vector2i, tile: TileDefinition?): IWorldChunkTuple<This, ChunkType> {
		val chunk = computeIfAbsentInternal(ChunkPos.fromTilePosition(pos))
		chunk.chunk.background[ChunkPos.normalizeCoordinate(pos.x), ChunkPos.normalizeCoordinate(pos.y)] = tile
		return chunk
	}

	protected open fun collectInternal(boundingBox: AABBi): List<IMutableWorldChunkTuple<This, ChunkType>> {
		val output = ArrayList<IMutableWorldChunkTuple<This, ChunkType>>()

		for (pos in boundingBox.chunkPositions) {
			val chunk = getChunkInternal(pos)

			if (chunk != null) {
				output.add(chunk)
			}
		}

		return output
	}

	/**
	 * Возвращает все чанки, которые пересекаются с заданным [boundingBox]
	 */
	open fun collect(boundingBox: AABBi): List<IWorldChunkTuple<This, ChunkType>> {
		val output = ArrayList<IWorldChunkTuple<This, ChunkType>>()

		for (chunk in collectInternal(boundingBox)) {
			output.add(WorldChunkTuple(chunk))
		}

		return output
	}

	/**
	 * Производит проверку на пересечение [worldaabb] с геометрией мира при попытке пройти в [_deltaMovement]
	 */
	fun sweep(
		worldaabb: AABB,
		_deltaMovement: Vector2d,
		collisionResolution: CollisionResolution,
		delta: Double
	): WorldSweepResult {
		var deltaMovement = _deltaMovement

		var potentialAABB = worldaabb + deltaMovement
		var combined = worldaabb.combine(potentialAABB)

		val collected = collectInternal((combined).encasingChunkPosAABB())
			.map { it.chunk.foreground.collisionLayers() }
			.flatten()
			.sortedWith { o1, o2 ->
				val a = o1.distance(worldaabb)
				val b = o2.distance(worldaabb)

				if (a == b)
					return@sortedWith 0

				if (a > b)
					return@sortedWith 1

				return@sortedWith -1
			}

		var hitAnything = false
		var hitNormal = Vector2d.ZERO

		if (collected.isNotEmpty()) {
			for (iteration in 0 .. 100) {
				var collided = false

				for (aabb in collected) {
					// залез в блоки?
					/*if (potentialAABB.intersectWeak(aabb)) {
						val push = worldaabb.pushOutFrom(aabb)
						println("into tiles")

						if (push.length > 0.0) {
							if (iteration == 0) {
								collided = true
								continue
							} else {
								hitAnything = true
								deltaMovement -= push * delta * 5.0
								potentialAABB = worldaabb + deltaMovement
								combined = worldaabb.combine(potentialAABB)

								collided = true
								continue
							}
						}
					}*/

					// ранний тест (отсечение заведомо не пересекаемой геометрии)
					if (!aabb.intersect(combined)) {
						continue
					}

					// обычный тест коллизии
					val (normal, collisionTime) = worldaabb.sweep(aabb, deltaMovement)

					if (collisionTime != 1.0) {
						hitAnything = true
						collided = true
						hitNormal = normal

						if (weakDoubleZeroing(collisionTime) == 0.0) {
							// Мы ВПЛОТНУЮ стоим к другой геометрии (пересекаемся одной из осей)
							deltaMovement = Vector2d.ZERO
						} else {
							val remainingTime = 1.0 - collisionTime

							when (collisionResolution) {
								CollisionResolution.STOP -> {
									// val weak = weakDoubleZeroing(collisionTime, epsilon = EPSILON)

									for (i in 100 downTo 1) {
										val move = deltaMovement * collisionTime * (i / 100.0)

										// сегодня, в программе
										// погрешность плавающей запятой
										if (!(worldaabb + move).intersect(aabb)) {
											deltaMovement = move
											break
										} // ЭтО чО вАщЕ ТакОе?
									}
								}

								CollisionResolution.PUSH -> {
									var dot = deltaMovement.pseudoDot(normal)
									val magnitude = deltaMovement.length * weakDoubleZeroing(remainingTime, epsilon = EPSILON)

									if (dot > 0.0) {
										dot = 1.0
									} else {
										dot = -1.0
									}

									deltaMovement = Vector2d(dot * normal.y * magnitude, dot * normal.x * magnitude)
								}

								CollisionResolution.SLIDE -> {
									val dot = deltaMovement.pseudoDot(normal) * remainingTime

									if (weakDoubleZeroing(dot) != 0.0) {
										deltaMovement = Vector2d(dot * normal.y, dot * normal.x)
									} else {
										// Граничный случай: Направления идеально противоположны.
										// Обрабатываем как CollisionResolution.STOP

										// val weak = weakDoubleZeroing(collisionTime, epsilon = EPSILON)

										for (i in 100 downTo 1) {
											val move = deltaMovement * collisionTime * (i / 100.0)

											// сегодня, в программе
											// погрешность плавающей запятой
											if (!(worldaabb + move).intersect(aabb)) {
												deltaMovement = move
												break
											} // ЭтО чО вАщЕ ТакОе?
										}
									}
								}

								CollisionResolution.BOUNCE -> {
									deltaMovement *= weakDoubleZeroing(remainingTime, epsilon = EPSILON)

									if (normal.x.absoluteValue > 0.00001 && normal.y.absoluteValue > 0.00001) {
										deltaMovement *= Vector2d.NEGATIVE_XY
									} else if (normal.x.absoluteValue > 0.00001) {
										deltaMovement *= Vector2d.NEGATIVE_X
									} else if (normal.y.absoluteValue > 0.00001) {
										deltaMovement *= Vector2d.NEGATIVE_Y
									}
								}
							}
						}

						potentialAABB = worldaabb + deltaMovement
					}
				}

				if (!collided) {
					//println("resolved in $iteration")
					break
				}
			}
		}

		return WorldSweepResult(
			hitAnything = hitAnything,
			hitNormal = hitNormal,
			collisionTime = worldaabb.centre.distance(deltaMovement) / worldaabb.centre.distance(_deltaMovement),
			hitPosition = deltaMovement,
		)
	}

	/**
	 * Возвращает, застрянет ли сущность будет с коллизией [worldaabb]
	 */
	fun isSpaceEmptyFromTiles(worldaabb: AABB): Boolean {
		val collected = collectInternal(worldaabb.encasingChunkPosAABB())
			.map { it.chunk.foreground.collisionLayers() }
			.flatten()
			.sortedWith { o1, o2 ->
				val a = o1.distance(worldaabb)
				val b = o2.distance(worldaabb)

				if (a == b)
					return@sortedWith 0

				if (a > b)
					return@sortedWith 1

				return@sortedWith -1
			}

		if (collected.isEmpty()) {
			return true
		}

		for (aabb in collected) {
			if (aabb.intersectWeak(worldaabb)) {
				return false
			}
		}

		return true
	}
}
