package ru.dbotthepony.kstarbound.world.entities

import ru.dbotthepony.kstarbound.math.lerp
import ru.dbotthepony.kvector.util2d.AABB
import ru.dbotthepony.kvector.vector.ndouble.Vector2d

enum class CollisionResolution {
	STOP,
	BOUNCE,
	PUSH,
	SLIDE,
}

abstract class MovementController<T : IEntity>(val entity: T) {
	val world = entity.world
	var pos by entity::pos
	var rotation by entity::rotation

	open val mass = 1.0

	// наследуемые свойства
	open val affectedByGravity = true
	open val collisionResolution = CollisionResolution.STOP

	var velocity = Vector2d()

	/**
	 * Касается ли AABB сущности земли
	 *
	 * Данный флаг выставляется при обработке скорости, если данный флаг не будет выставлен
	 * правильно, то сущность будет иметь очень плохое движение в стороны
	 *
	 * Так же от него зависит то, может ли сущность двигаться, если она не парит
	 *
	 * Если сущность касается земли, то на неё не действует гравитация
	 */
	var onGround = false
		protected set(value) {
			field = value
			nextOnGroundUpdate = world.timer + 0.1
		}

	/**
	 * Текущий AABB этого Movement Controller
	 *
	 * Это может быть как и статичное значение (для данного типа сущности), так и динамичное
	 * (к примеру, присевший игрок)
	 *
	 * Данное значение, хоть и является val, НЕ ЯВЛЯЕТСЯ КОНСТАНТОЙ!
	 */
	abstract val currentAABB: AABB

	/**
	 * Текущий AABB в отображении на мировые координаты
	 */
	val worldAABB: AABB get() = currentAABB + pos

	protected var nextOnGroundUpdate = 0.0

	protected fun sweepRelative(velocity: Vector2d, delta: Double, collisionResolution: CollisionResolution = this.collisionResolution) = world.sweep(worldAABB, velocity, collisionResolution, delta)
	protected fun sweepAbsolute(from: Vector2d, velocity: Vector2d, delta: Double, collisionResolution: CollisionResolution = this.collisionResolution) = world.sweep(currentAABB + from, velocity, collisionResolution, delta)
	protected fun isSpaceOpen(relative: Vector2d, delta: Double) = !sweepRelative(relative, delta).hitAnything

	fun dropToFloor() {
		val sweep = sweepRelative(DROP_TO_FLOOR, 1.0, CollisionResolution.STOP)

		if (!sweep.hitAnything)
			return

		pos += sweep.hitPosition
	}

	var groundNormal = Vector2d.ZERO
		protected set

	protected open fun propagateVelocity(delta: Double) {
		if (velocity.length == 0.0)
			return

		val sweep = sweepRelative(velocity * delta, delta)
		this.velocity = sweep.hitPosition / delta
		this.pos += this.velocity * delta

		if (nextOnGroundUpdate <= world.timer || !onGround) {
			onGround = sweep.hitNormal.dot(world.gravity.normalized) <= -0.98
			groundNormal = sweep.hitNormal

			if (!onGround) {
				val sweepGround = sweepRelative(world.gravity * delta, delta)
				onGround = sweepGround.hitAnything && sweepGround.hitNormal.dot(world.gravity.normalized) <= -0.98
				groundNormal = sweepGround.hitNormal
			}
		}
	}

	protected open fun thinkGravity(delta: Double) {
		velocity += world.gravity * delta
	}

	protected open fun thinkFriction(delta: Double) {
		velocity *= Vector2d(lerp(delta, 1.0, 0.01), 1.0)
	}

	open fun thinkPhysics(delta: Double) {
		if (!onGround && affectedByGravity)
			thinkGravity(delta)

		propagateVelocity(delta)

		if (affectedByGravity && onGround)
			thinkFriction(delta)
	}

	protected open fun onTouchSurface(velocity: Vector2d, normal: Vector2d) {
		entity.onTouchSurface(velocity, normal)
	}

	companion object {
		private val DROP_TO_FLOOR = Vector2d(y = -1_000.0)
	}
}

/**
 * MovementController который ничего не делает (прям совсем)
 */
class DummyMovementController(entity: Entity) : MovementController<Entity>(entity) {
	override val currentAABB = DUMMY_AABB
	override val affectedByGravity = false

	override fun propagateVelocity(delta: Double) {
		// no-op
	}

	override fun thinkGravity(delta: Double) {
		// no-op
	}

	override fun thinkFriction(delta: Double) {
		// no-op
	}

	companion object {
		private val DUMMY_AABB = AABB.rectangle(Vector2d.ZERO, 0.1, 0.1)
	}
}
